/** @file
 * @brief Topology description API
 * @ingroup Topology
 */

/** @defgroup Topology API: Grid Topology management
 * @ingroup Topology
 */

#ifndef PADICO_TOPOLOGY_H
#define PADICO_TOPOLOGY_H

#ifdef __cplusplus
extern "C" {
#endif

#include <Padico/Puk.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <errno.h>
#include <netinet/tcp.h>

/* ********************************************************* */
/* *** Topology *************************************************** */

/* Definitions for topology management:
 *  -- host: a machine, identified by a hostname.
 *  -- node: a computing node (ie. process). There may 
 *     be several nodes per host. A node is identified 
 *     by a nodename.
 *  -- network: a list (order matters) of nodes or hosts with 
 *     bindings between network and nodes/hosts.
 *     Binding address may be rank-based (nodes) or
 *     in_addr-based (hosts).
 *
 */


/** @addtogroup Topology
 * @{
 */

/* *** Types */

#define PADICO_TOPO_UUID_SIZE 32
typedef const struct topo_uuid_s*padico_topo_uuid_t;
typedef       struct topo_host_s*padico_na_host_t;
typedef       struct topo_node_s*padico_na_node_t;
typedef       struct topo_network_s*padico_topo_network_t;
typedef       struct topo_binding_s*padico_topo_binding_t;
typedef const struct topo_hwaddr_s*padico_topo_hwaddr_t;
typedef const struct topo_address_s*padico_topo_address_t;
typedef       struct topo_filter_s*padico_topo_filter_t;
typedef       struct topo_service_s*padico_topo_service_t;

PUK_VECT_TYPE(padico_na_node,    padico_na_node_t);
PUK_VECT_TYPE(padico_na_network, padico_topo_network_t);
PUK_VECT_TYPE(padico_na_host,    padico_na_host_t);

/* *** Addresses and bindings */

typedef enum
  { 
    PADICO_NA_SCOPE_NONE = 0x00,
    PADICO_NA_SCOPE_NODE = 0x01,
    PADICO_NA_SCOPE_HOST = 0x02,
    PADICO_NA_SCOPE_ALL  = 0xFF
  } padico_na_binding_scope_t;
  
typedef enum
  {
    TOPO_ADDR_VISIBILITY_NONE      = 0,
    TOPO_ADDR_VISIBILITY_LOOPBACK  = 1, /**< loopback address- only local node/host can see it */
    TOPO_ADDR_VISIBILITY_LINKLOCAL = 2, /**< link-local address- nodes on the same physical link can see it */
    TOPO_ADDR_VISIBILITY_PRIVATE   = 3, /**< private address- some node/host cannot see it */
    TOPO_ADDR_VISIBILITY_PUBLIC    = 4, /**< public address */
  } padico_topo_addr_visibility_t; 

typedef enum
  {
    _TOPO_FILTER_NONE     = 0x00,
    TOPO_FILTER_BLOCK_IN  = 0x01,
    TOPO_FILTER_BLOCK_OUT = 0x02,
    TOPO_FILTER_ALLOW_IN  = 0x04,
    TOPO_FILTER_ALLOW_OUT = 0x08,
    _TOPO_FILTER_MAX
  } padico_topo_filter_policy_t;
#define TOPO_FILTER_MASK_IN    (TOPO_FILTER_BLOCK_IN  | TOPO_FILTER_ALLOW_IN)
#define TOPO_FILTER_MASK_OUT   (TOPO_FILTER_BLOCK_OUT | TOPO_FILTER_ALLOW_OUT)
#define TOPO_FILTER_MASK_BLOCK (TOPO_FILTER_BLOCK_IN  | TOPO_FILTER_BLOCK_OUT)
#define TOPO_FILTER_MASK_ALLOW (TOPO_FILTER_ALLOW_IN  | TOPO_FILTER_ALLOW_AOUT)
  
#define PORT_ANY 0
  
  /* *** Host ********************************************** */

  /**  Create new host
   */
  padico_na_host_t     padico_topo_host_create(const char*hostname);

  /**  Add a name to a host;
   * @param resolvable whether the host is DNS-resolvable (0: no; 1: yes; -1: don't known => detect).
   */
  void padico_topo_host_add_name(padico_na_host_t host, const char*name, int resolvable);

  /**  Add a service to a host
   */
  void padico_topo_host_add_service(padico_na_host_t host, const char*service_name, uint16_t service_port);

  /**  Add a property to a host
   */
  void padico_topo_host_add_property(padico_na_host_t host, const char*label, const char*name);

  /** Get a host description record from an inet address (create if it didn't exist)
   */
  padico_na_host_t      padico_na_host_resolve_byinaddr(const struct in_addr*addr);
  
  /** Get a host from an inet address (return NULL if host doesn't exist).
   */
  padico_na_host_t      padico_na_host_getbyinaddr(const struct in_addr*addr);

  /** Get a host description record from an inet6 address (create if it didn't exist) 
   */
  padico_na_host_t      padico_na_host_resolve_byin6addr(const struct in6_addr*addr);

  /** Get a host from an inet6 address (return NULL if host doesn't exist).
   */
  padico_na_host_t      padico_na_host_getbyin6addr(const struct in6_addr*addr);

  /** Get a host description record (lookup in cache, create if it didn't exist)
   */
  padico_na_host_t      padico_na_host_resolvebyname(const char*hostname);

  /** Get the canonical name of a host
   */
  const char*           padico_na_host_getname(padico_na_host_t host);

  /** Get the canonical name of a host on a network 
   */
  const char*           padico_na_host_getname2(padico_na_host_t host, padico_topo_network_t network);

  /** Get the canonical IP address of a host
   */
  const struct in_addr* padico_na_host_getaddr(padico_na_host_t host);

  /** Get the canonical IP address of a host on a network
   */
  const struct in_addr* padico_na_host_getaddr2(padico_na_host_t host, padico_topo_network_t network);

  /** Get the binding of a host on a network.
   */
  padico_topo_binding_t padico_na_host_getbinding(padico_na_host_t host, padico_topo_network_t network);

  /** Get the nodes on a host
   */
  padico_na_node_vect_t padico_na_host_getnodes(padico_na_host_t host);

  /** Get the networks a host is connected to
   */
  void padico_na_host_getnetworks(padico_na_host_t host, padico_na_network_vect_t networks);

  /** Get the service port of the service on a host
   */
  int padico_na_host_getservice(padico_na_host_t host, char*service_name);
  
  /** Get property on a host
   */
  const char*padico_na_host_getproperty(padico_na_host_t host, const char*label);

  /** Verify if the port is filtered output between host and network
   */
  int padico_na_host_can_output(padico_na_host_t host, padico_topo_network_t network, uint16_t port, int protocol);

  /** Verify if the port is filtered input between host and network
   */
  int padico_na_host_can_input(padico_na_host_t host, padico_topo_network_t network, uint16_t port, int protocol);  

  int padico_topo_network_can_input(padico_topo_network_t subnet, padico_topo_network_t network, uint16_t port, int protocol);
  int padico_topo_network_can_output(padico_topo_network_t subnet, padico_topo_network_t network, uint16_t port, int protocol);

  /** Declare a new binding for a host
   */
  padico_topo_binding_t padico_na_host_bind(padico_na_host_t host, padico_topo_network_t network, const void*address);

  /** Get known hosts*/
  padico_na_host_vect_t padico_na_gethosts();

  /** Get the first node of a host
   * @deprecated
   */
  static inline padico_na_node_t padico_na_host_getnode(padico_na_host_t host)
  {
    const padico_na_node_vect_t nodes = padico_na_host_getnodes(host);
    return padico_na_node_vect_empty(nodes)?NULL:*(padico_na_node_vect_begin(nodes));
  }

  /* *** Nodes ********************************************* */
  
  /** node lookup, by node name
   */
  padico_na_node_t      padico_na_getnodebyname  (const char*nodename);
  
  /** node lookup, by UUID
   */
  padico_na_node_t      padico_na_getnodebyuuid  (padico_topo_uuid_t uuid);

  /** Get the UUID of a node
   */
  padico_topo_uuid_t    padico_na_node_getuuid   (padico_na_node_t node);

  /** Get the canonical name of a node
   */
  const char*           padico_na_node_getname   (padico_na_node_t node);
  
  /** Get the host hosting a node
   */
  padico_na_host_t      padico_na_node_gethost   (padico_na_node_t node);

  /** Get the address of the node in the given network
   */
  padico_topo_address_t padico_na_node_getaddress(padico_na_node_t node, padico_topo_network_t network);

  /** Get the boot session ID
   */
  const char*           padico_na_node_getsession(padico_na_node_t node);

  /** Get the networks a node is connected to
   */
  void padico_na_node_getnetworks(padico_na_node_t node, padico_na_network_vect_t networks);

  /** Remove a disconnected node from topology
   */
  void padico_na_node_disconnect(padico_na_node_t node);

  /** Declare a new binding for a node
   */
  padico_topo_binding_t padico_na_node_bind(padico_na_node_t node, padico_topo_network_t network, const void*address);


  /* *** Networks ****************************************** */

  padico_topo_network_t padico_topo_network_create(const char*id, padico_na_binding_scope_t scope, size_t addr_size, int family);

  padico_topo_network_t padico_na_network_getbyid(const char*network_id);

  /** Get networks of a given address family (e.g. AF_INET, AF_INET6, etc.)
   */
  padico_na_network_vect_t padico_topo_networks_getbyfamily(int family);

  /** tests whether the given node belong to the network
   */
  int                   padico_na_network_contains_node(padico_topo_network_t network, padico_na_node_t node);
  const char*           padico_na_network_getname(padico_topo_network_t network);
  int                   padico_na_network_size   (padico_topo_network_t network);
  /** Get a host description record from an address
   */
  padico_na_host_t      padico_na_network_gethostbyaddr(padico_topo_network_t, const void*);

  /** Get a node descriptor from an address
   */
  padico_na_node_t      padico_na_network_getnodebyaddr(padico_topo_network_t network, const void*addr);

  void                  padico_na_network_gethosts(padico_topo_network_t n, padico_na_host_vect_t hosts);

  void                  padico_na_network_get_common_host(padico_topo_network_t n1, padico_topo_network_t n2, padico_na_host_vect_t hosts);

  void                  padico_na_network_set_perf(padico_topo_network_t n, int latency, int bandwidth);

  /* *** serialization ************************************* */

  padico_string_t padico_na_host_serialize(padico_na_host_t host);
  padico_string_t padico_na_node_serialize(padico_na_node_t node);
  padico_string_t padico_na_network_serialize(padico_topo_network_t net);
  padico_string_t padico_na_topology_serialize(padico_na_binding_scope_t scope);


  /* *** local info *****************************************/

  padico_na_node_t padico_na_getlocalnode(void);

#define padico_na_nodename()     padico_na_node_getname(padico_na_getlocalnode())
#define padico_na_nodealias()    padico_na_node_getalias(padico_na_getlocalnode())
#define padico_na_getlocalhost() padico_na_node_gethost(padico_na_getlocalnode())
#define padico_na_hostname()     padico_na_host_getname(padico_na_getlocalhost())


  /* *** session ********************************************/

  static inline padico_topo_network_t padico_na_session_getnetwork(const char*session_id)
  {
    padico_string_t network_name = padico_string_new();
    padico_string_printf(network_name, "Session-[%s]", session_id);
    padico_topo_network_t network = padico_na_network_getbyid(padico_string_get(network_name));
    padico_string_delete(network_name);
    return network;
  }
  static inline padico_na_node_t padico_na_session_getnodebyrank(const char*session_id, int rank)
  {
    padico_topo_network_t network = padico_na_session_getnetwork(session_id);
    padico_na_node_t node = padico_na_network_getnodebyaddr(network, &rank);
    return node;
  }

  /* *** hwaddr ********************************************/

  padico_topo_hwaddr_t padico_topo_hwaddr_create(const char*name, int index, int type, size_t len, const void*addr);

  void padico_topo_hwaddr_get(padico_topo_hwaddr_t hwaddr, const char**name, int*index, int*type, size_t*len, const void**addr);
  
  void padico_topo_hwaddr_add(padico_na_host_t host, padico_topo_hwaddr_t*_hwaddr);
  
  /* *** bindings ********************************************/
  
  void padico_topo_binding_add_addr(padico_topo_binding_t binding, const void*addr);

  void padico_topo_binding_add_hwaddr(padico_topo_binding_t binding, padico_topo_hwaddr_t hwaddr);
    
  void padico_topo_binding_add_gw_addr(padico_topo_binding_t binding, struct in_addr*inaddr);

  void padico_topo_binding_add_hostname(padico_topo_binding_t binding, const char*name);

  void padico_topo_binding_add_route(padico_topo_binding_t binding, padico_topo_network_t dest, const void*gwaddr);
  
  void padico_topo_binding_add_filter(padico_topo_binding_t binding, padico_topo_filter_policy_t policy, uint16_t port, int protocol);

  void padico_topo_binding_add_perf(padico_topo_binding_t binding, int latency, int bandwidth);

  padico_topo_hwaddr_t padico_topo_binding_gethwaddr(padico_topo_binding_t binding);

  /* *** address visibility ********************************/
  
  padico_topo_addr_visibility_t padico_topo_inet_visibility(in_addr_t _inaddr);

  padico_topo_addr_visibility_t padico_topo_inet6_visibility(const struct in6_addr*in6addr);

/* *** backward compatibility */
#include <Padico/Module.h>
#define padico_na_size()  atoi(padico_getattr("PADICO_BOOT_SIZE")?:"1")
#define padico_na_rank()  atoi(padico_getattr("PADICO_BOOT_RANK")?:"0")



/* ********************************************************* */
/* *** Padico addresses as sockaddr ************************ */

/* padico_sockaddr (struct sockaddr_pa) define a new AF_PADICO
 * address space for Padico-to-Padico communications with
 * non-interoperable methods.
 */

typedef union 
{
  char bytes[4];
  uint32_t integer;
} padico_address_kind_t;

#define AF_PADICO (AF_MAX + 1) /* just to make sure that AF_PADICO addresses 
				  are never leaked into the OS... */
#define PF_PADICO AF_PADICO

/* make uint32_t from bytes */
#define AF_PADICO_MAGIC(B1, B2, B3, B4) \
  (((padico_address_kind_t){ .bytes = { B1, B2, B3, B4 } }).integer)
/* make address_kind_t from uint32_t */
#define AF_PADICO_MAKE(I) \
  ((padico_address_kind_t){ .integer = (I) })
/* make bytes from uint32_t */
#define AF_PADICO_REVERSE(I) \
  ((AF_PADICO_MAKE(I)).bytes)

/* embed an AF_INET address into an AF_PADICO */
#define AF_PADICO_INET AF_PADICO_MAGIC('I', 'N', 'E', 'T')
/* embed an AF_UNIX address into an AF_PADICO */
#define AF_PADICO_UNIX AF_PADICO_MAGIC('U', 'N', 'I', 'X')
/* requests an autobind (only suitable as *input* for bind(), should not be returned by bind()). */
#define AF_PADICO_AUTO AF_PADICO_MAGIC('A', 'U', 'T', 'O')

#define AF_PADICO_PAYLOAD 64

struct sockaddr_pa
{
  sa_family_t   spa_family; /* =AF_PADICO, host byte order */
  uint32_t      spa_magic;  /* magic number for sub-family, host byte order */
  unsigned char spa_payload[AF_PADICO_PAYLOAD];
};

/* Don't inline this code in C++ */
#ifndef __cplusplus

static inline void padico_sockaddr_init(struct sockaddr_pa*addr, socklen_t*addrlen, int magic)
{
  assert(addr && addrlen && *addrlen >= sizeof(struct sockaddr_pa));
  memset(addr->spa_payload, 0, AF_PADICO_PAYLOAD); /* reset with 0s; better for debug! */
  addr->spa_family = AF_PADICO;
  addr->spa_magic  = magic;
}
#define padico_sockaddr_payload(TYPE, ADDRESS) \
  ( assert(sizeof(TYPE) <= AF_PADICO_PAYLOAD), \
    (TYPE*)(((struct sockaddr_pa*)(ADDRESS))->spa_payload) )
#define padico_sockaddr_magic(ADDRESS) \
  ( assert(((struct sockaddr_pa*)(ADDRESS))->spa_family == AF_PADICO),	\
    ((struct sockaddr_pa*)(ADDRESS))->spa_magic )

static inline int padico_sockaddr_check(const struct sockaddr_pa*addr, int magic)
{
  const struct sockaddr*saddr = (struct sockaddr*)addr;
  if(saddr->sa_family != AF_PADICO)
    {
#ifndef NDEBUG
      padico_warning("bad address family- sa_family=%d; expected=%d (AF_PADICO).\n",
		     saddr->sa_family, AF_PADICO);
#endif
      return EAFNOSUPPORT;
    }
  else if(addr->spa_magic != magic)
    {
#ifndef NDEBUG
      const padico_address_kind_t m = AF_PADICO_MAKE(addr->spa_magic);
      const padico_address_kind_t e = AF_PADICO_MAKE(magic);
      padico_warning("bad Padico address magic- spa_magic=%d (%c%c%c%c); expected=%d (%c%c%c%c).\n",
		     addr->spa_magic, m.bytes[0], m.bytes[1], m.bytes[2], m.bytes[3],
		     magic,           e.bytes[0], e.bytes[1], e.bytes[2], e.bytes[3]);
#endif
      return EADDRNOTAVAIL;
    }
  return 0;
}
#endif /* __cplusplus */


/* ********************************************************* */
/* *** Padico vaddr **************************************** */

/* Padico vaddr (padico_vaddr_t) is a generic address used for
 * transmissions. It is mainly a packed, self-contained and 
 * serializable form of struct sockaddr.
 */

struct padico_vaddr_s
{
  struct sockaddr*addr;
  socklen_t addrlen;
  int proto;
};

typedef struct padico_vaddr_s*padico_vaddr_t;

padico_string_t padico_vaddr_marshall(const struct padico_vaddr_s*);
padico_vaddr_t  padico_vaddr_unmarshall(const char*);

/* C++ doesn't like inlined GNU extensions */
#ifndef __cplusplus
static inline padico_vaddr_t padico_vaddr_new(struct sockaddr*addr, socklen_t addrlen, int proto)
{
  padico_vaddr_t vaddr = padico_malloc(sizeof(struct padico_vaddr_s));
  *vaddr = (struct padico_vaddr_s)
    {
      .addr    = addr,
      .addrlen = addrlen,
      .proto   = proto
    };
  return vaddr;
}
static inline void padico_vaddr_delete(padico_vaddr_t vaddr)
{
  padico_free(vaddr);
}
#endif /* __cplusplus */



/** @} */

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* PADICO_TOPOLOGY_H */
