/** @file
 ** @brief Base types and Puk management code.
 * @ingroup Puk
 */

/* ** valgrind integration ********************************* */

#if defined(PUK_DEBUG) && defined(PUK_HAVE_VALGRIND_H)
#include <valgrind/valgrind.h>
#include <valgrind/memcheck.h>
#define PUK_VG_CREATE_MEMPOOL(pool, rz, zeroed)  VALGRIND_CREATE_MEMPOOL(pool, rz, zeroed)
#define PUK_VG_DESTROY_MEMPOOL(pool)             VALGRIND_DESTROY_MEMPOOL(pool)
#define PUK_VG_MEMPOOL_ALLOC(pool, addr, size)   VALGRIND_MEMPOOL_ALLOC(pool, addr, size)
#define PUK_VG_MEMPOOL_FREE(pool, addr)          VALGRIND_MEMPOOL_FREE(pool, addr)
#define PUK_VG_MAKE_MEM_NOACCESS(addr, len)      VALGRIND_MAKE_MEM_NOACCESS(addr, len)
#define PUK_VG_STACK_REGISTER(start, end)        VALGRIND_STACK_REGISTER(start, end)
#define PUK_VG_STACK_DEREGISTER(id)              VALGRIND_STACK_DEREGISTER(id)
#define PUK_VG_RUNNING_ON_VALGRIND               RUNNING_ON_VALGRIND
#else /* PUK_DEBUG && HAVE_VALGRIND_H */
#define PUK_VG_NOOP ((void)0)
#define PUK_VG_CREATE_MEMPOOL(pool, rz, zeroed)  PUK_VG_NOOP
#define PUK_VG_DESTROY_MEMPOOL(pool)             PUK_VG_NOOP
#define PUK_VG_MEMPOOL_ALLOC(pool, addr, size)   PUK_VG_NOOP
#define PUK_VG_MEMPOOL_FREE(pool, addr)          PUK_VG_NOOP
#define PUK_VG_MAKE_MEM_NOACCESS(addr, len)      PUK_VG_NOOP
#define PUK_VG_STACK_REGISTER(start, end)        PUK_VG_NOOP
#define PUK_VG_STACK_DEREGISTER(id)              PUK_VG_NOOP
#define PUK_VG_RUNNING_ON_VALGRIND               0
#endif /* PUK_DEBUG && HAVE_VALGRIND_H */


/* *** Typing ********************************************** */

/* use these macros to display size_t and ssize_t in
 * a portable (32/64 bits) way */
#define puk_ulong(N) ((unsigned long)(N))
#define puk_slong(N) ((long)(N))
/** convert a long to pointer */
#define puk_long2ptr(N) ((void*)(long)N)
/** convert a pointer to a long */
#define puk_ptr2long(P) ((long)(P))


/* *** Linking ********************************************* */

#define __PUK_SYM_INTERNAL __attribute__ ((visibility("internal")))


  /** Initializes Puk */
  extern void padico_puk_init(int argc, char**argv);

  /** add the given path to Puk global search path */
  extern void padico_puk_add_path(const char*path);

  /** Start Puk bootstrap module */
  extern void padico_puk_start(void);

  /** Shutdowns Puk */
  extern void padico_puk_shutdown(void);

  /** Tests whether Puk is already initialized */
  extern int padico_puk_initialized(void);

  /** Tests whether Puk is finalized (or shutdown in progress) */
  extern int padico_puk_finalized(void);


  /* ** register locking handlers for Puk shared structures */
  void puk_lock_set_handlers(void(*_acquire)(void), void(*_release)(void));

  /* Data structures: lists, vectors, hashtables */

#define PUK_AUTO_TYPE(ENAME) \
  PUK_AUTO_TYPE_DECLARE(ENAME) \
  PUK_AUTO_TYPE_CREATE(ENAME)

#define PUK_AUTO_TYPE_DECLARE(ENAME) \
  typedef struct ENAME##_s * ENAME##_t;

#define PUK_AUTO_TYPE_CREATE(ENAME) \
  void ENAME##_ctor(ENAME##_t); \
  void ENAME##_dtor(ENAME##_t); \
  /** @internal */ \
  static inline ENAME##_t ENAME##_new(void) \
  { ENAME##_t obj=(ENAME##_t)padico_malloc(sizeof(struct ENAME##_s)); ENAME##_ctor(obj); return obj; } \
  /** @internal */ \
  static inline void ENAME##_delete(ENAME##_t obj) \
  { ENAME##_dtor(obj); padico_free(obj); }


  /** Sets the value of an environment variable. */
  void padico_setenv(const char *var, const char *val);

  /** Returns the value of an environment variable. */
  const char*padico_getenv(const char*var);

  /** Remove padico environment variables*/
  void padico_cleanenv(void);

  /** whether we are running on mckernel */
  int puk_mckernel(void);

  const char*padico_hostname(void);

  struct in_addr puk_inet_getaddr(void);

  /** Abort the processus with a fatal error.
   * @param format Error message in "printf" format
   * @warning Use with care! It aborts the *whole process*, not only the thread/module.
   */
  void padico_fatal(const char*format, ...) __attribute__ ((noreturn))  __attribute__ ((format (printf, 1, 2)));

  void padico_warning(const char*format, ...) __attribute__ ((format (printf, 1, 2)));

  void padico_print(const char*format, ...)  __attribute__ ((format (printf, 1, 2)));


  /* padico_malloc & co. Everything is managed in PukABI, so nothing to do now */
#define padico_malloc  malloc
#define padico_free    free
#define padico_realloc realloc
#define padico_calloc  calloc
#define padico_strdup  strdup


/* ** base64 encoding, from puk-base64.c
 * usable standalone, without full Puk.
 */
char*puk_base64_encode(const void*bytes, int*len, void*(*alloc)(unsigned long));
void*puk_base64_decode(const char*text,  int*len, void*(*alloc)(unsigned long));
