/** @file
 * @brief interface for timing
 */

#ifndef PUK_TIMING_H
#define PUK_TIMING_H

#include "Puk-config.h"
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>

#if defined(MCKERNEL) && defined (PUK_HAVE_CLOCK_GETTIME)
#undef PUK_HAVE_CLOCK_GETTIME
#endif


#if defined(PUK_HAVE_CLOCK_GETTIME)

/* ** timing with clock_gettime **************************** */

#define PUK_TIMING_USE_CLOCK_GETTIME

typedef struct timespec puk_tick_t;

#ifdef CLOCK_MONOTONIC_RAW
#  define CLOCK_TYPE CLOCK_MONOTONIC_RAW
#else
#  define CLOCK_TYPE CLOCK_MONOTONIC
#endif

#define PUK_GET_TICK(t)				\
	clock_gettime(CLOCK_TYPE, &(t))
#define PUK_TICK_RAW_DIFF(start, end)		\
        puk_tick_raw_diff(&(start), &(end))

static inline puk_tick_t puk_tick_raw_diff(const puk_tick_t*const t1, const puk_tick_t*const t2)
{
  puk_tick_t diff;
  if(t2->tv_nsec > t1->tv_nsec)
    {
      diff.tv_sec  = t2->tv_sec - t1->tv_sec;
      diff.tv_nsec = t2->tv_nsec - t1->tv_nsec;
    }
  else
    {
      diff.tv_sec  = t2->tv_sec - t1->tv_sec - 1;
      diff.tv_nsec = t2->tv_nsec - t1->tv_nsec + 1000000000L;
    }
  return diff;
}

#elif defined (PUK_HAVE_MACH_ABSOLUTE_TIME)

/* ** timing with mach_absolute_time *********************** */

#define PUK_TIMING_USE_MACH_ABSOLUTE_TIME

#include <mach/mach_time.h>

typedef long long puk_tick_t;

#define PUK_GET_TICK(t)				\
	t = mach_absolute_time()
#define PUK_TICK_RAW_DIFF(start, end)		\
	((end) - (start))


#elif defined(__x86_64__)

/* ** timing with rdtsc ************************************ */

#define PUK_TIMING_USE_RDTSC

typedef unsigned long long puk_tick_t;

/* old 32 bits x86 code */
/*
#define PUK_GET_TICK(t)				\
  __asm__ volatile("rdtsc" : "=A" (t))
*/
#define PUK_GET_TICK(t) do { \
     unsigned int __a,__d; \
     asm volatile("rdtsc" : "=a" (__a), "=d" (__d)); \
     (t) = ((unsigned long)__a) | (((unsigned long)__d)<<32); }	\
  while(0)
#define PUK_TICK_RAW_DIFF(t1, t2) \
   ((t2) - (t1))


#else

/* ** timing with gettimeofday ***************************** */

#define PUK_TIMING_USE_GETTIMEOFDAY

typedef struct timeval puk_tick_t;

#define PUK_GET_TICK(t)				\
	gettimeofday(&(t), NULL)
#define PUK_TICK_RAW_DIFF(start, end)		\
        puk_tick_raw_diff(&(start), &(end))

static inline puk_tick_t puk_tick_raw_diff(const puk_tick_t*const t1, const puk_tick_t*const t2)
{
  puk_tick_t diff;
  if(t2->tv_usec > t1->tv_usec)
    {
      diff.tv_sec  = t2->tv_sec - t1->tv_sec;
      diff.tv_usec = t2->tv_usec - t1->tv_usec;
    }
  else
    {
      diff.tv_sec  = t2->tv_sec - t1->tv_sec - 1;
      diff.tv_usec = t2->tv_usec - t1->tv_usec + 1000000L;
    }
  return diff;
}

#endif

extern double puk_ticks2delay(const puk_tick_t *t1, const puk_tick_t *t2);

extern double puk_tick2usec(puk_tick_t t);

/** compute delay between to ticks, result given in microsecond as a double
 */
#define PUK_TIMING_DELAY(start, end) puk_ticks2delay(&(start), &(end))

/* ** compatibility layer for padico_timing_* API */

typedef puk_tick_t padico_timing_t;
static inline void padico_timing_get_tick(padico_timing_t*t)
{
  PUK_GET_TICK(*t);
}
static inline double padico_timing_diff_usec(padico_timing_t*t1, padico_timing_t*t2)
{
  return PUK_TIMING_DELAY(*t1, *t2);
}

#endif /* PUK_TIMING_H */
