/** @file
 * @brief Provides "binary" module driver
 * @ingroup Puk
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "Puk-internals.h"
#include "Module.h"

PADICO_MODULE_HOOK(Puk);

/* ********************************************************* */

static padico_rc_t unit_binary_load  (puk_unit_t unit);
static padico_rc_t unit_binary_start (puk_unit_t unit, puk_job_t job);
static padico_rc_t unit_binary_unload(puk_unit_t unit);

const static struct padico_loader_s binary_driver =
{
  .name   = "binary",
  .load   = &unit_binary_load,
  .start  = &unit_binary_start,
  .unload = &unit_binary_unload
};

__PUK_SYM_INTERNAL
void padico_puk_binary_init(void)
{
  puk_component_declare("PadicoLoader-binary", puk_component_provides("PadicoLoader", "loader", &binary_driver));
}

/* ********************************************************* */

/** a library handler, used as driver_specific for 'binary' */
struct puk_lib_s
{
  char*filename;  /**< library file name */
  void*dl_handle; /**< handle for dlopen */
};
PUK_AUTO_TYPE(puk_lib)
static padico_rc_t   puk_lib_open   (puk_lib_t* lib, const char*lib_name);
static void          puk_lib_setself(puk_lib_t lib, puk_mod_t mod);
static int           puk_lib_init   (puk_lib_t lib);
static puk_run_fun_t puk_lib_run_resolve (puk_lib_t lib);
static padico_rc_t   puk_lib_unload (puk_lib_t lib);

#ifndef RTLD_DEFAULT
#  define RTLD_DEFAULT NULL
#endif
#ifndef RTLD_NEXT
# define RTLD_NEXT ((void *) -1l)
#endif

#define PUK_DYLIB_DEFAULT_MODE   (RTLD_NOW | RTLD_GLOBAL)
#define PUK_DYLIB_DEFAULT_HANDLE  RTLD_DEFAULT
#define PUK_DYLIB_NEXT_HANDLE     RTLD_NEXT

static void*puk_dylib_open(const char *pathname, int mode);
static int  puk_dylib_close(void *handle);
static char*puk_dylib_error(void);
static void*puk_dylib_resolve(void*handle, const char*symbol);


/* ********************************************************* */

static padico_rc_t unit_binary_load(puk_unit_t unit)
{ 
  padico_rc_t rc = NULL;
  puk_lib_t lib = NULL;
  if(unit->driver_specific)
    {
      rc = padico_rc_error("binary driver: load error- unit **%s** is already loaded",
			   unit->name);
      return rc;
    }
  padico_string_t s_lib_file = padico_string_new();
  if(unit->mod->mod_path)
    {
      padico_string_printf(s_lib_file, "%s/", unit->mod->mod_path);
    }
  padico_string_catf(s_lib_file, "binary/%s", unit->name);
  rc = puk_lib_open(&lib, padico_string_get(s_lib_file));
  if(padico_rc_iserror(rc))
    {
      padico_warning("Puk: *%s* (%s) load error %s.\n", unit->name, padico_string_get(s_lib_file),
		     padico_rc_gettext(rc));
    }
  else
    {
      unit->driver_specific = (void*)lib;
      puk_lib_setself(lib, unit->mod);
      int err = puk_lib_init(lib);
      if(err)
	{
	  padico_warning("Puk: *%s* init error.\n", unit->name);
	  rc = padico_rc_error("binary driver: unit **%s** load error- module initialization rc=%d", 
			       unit->name, err);
	}
    }
  padico_string_delete(s_lib_file);
  return rc;
}

static padico_rc_t unit_binary_start(puk_unit_t unit, puk_job_t job)
{
  padico_rc_t rc = NULL; /* rc for the *start* operation (not for the job) */
  puk_lib_t lib  = (puk_lib_t)unit->driver_specific;
  if(lib != NULL)
    {
      puk_run_fun_t fun  = puk_lib_run_resolve(lib);
      if(fun != NULL)
	{
	  int i;
	  padico_out(50, "starting unit %s\n", unit->name);
	  
	  if(!job->argv)
	    {
	      job->argc = 1;
	      job->argv = padico_malloc(2 * sizeof(char*));
	      job->argv[0] = padico_strdup(unit->name);
	      job->argv[1] = NULL;
	    }
	  
	  padico_out(50, "argc = %d\n", job->argc);
	  for(i = 0; i < job->argc + 1; i++)
	    padico_out(50, " %d : %s\n", i, job->argv[i]);
	  
	  int rc = (*(fun))(job->argc, job->argv);
	  if(rc)
	    {
	      job->rc = padico_rc_new();
	      job->rc->rc  = rc;
	      job->rc->msg = padico_string_new();
	    }
	  else
	    {
	      job->rc = padico_rc_ok();
	    }
	  puk_job_notify(job);
	}
      else
	{
	  rc = padico_rc_error("binary driver: cannot start unit **%s** (not runnable)",
			       unit->name);
	}
    }
  else
    {
      rc = padico_rc_error("binary driver: cannot start unit **%s** (not properly initialized)",
			   unit->name);
    }
  return rc;
}

static padico_rc_t unit_binary_unload(puk_unit_t unit)
{
  const puk_lib_t lib = (puk_lib_t)unit->driver_specific;
  padico_rc_t rc = puk_lib_unload(lib);
  unit->driver_specific = NULL;
  return rc;
}


/* ***** Puk dylib ****************************************** */

static inline void puk_dylib_lock(void)
{
  puk_lock_acquire();
}
static inline void puk_dylib_unlock(void)
{
  puk_lock_release();
}

void*puk_dylib_open(const char *pathname, int mode)
{
  static int first = 1;
  if(PUK_VG_RUNNING_ON_VALGRIND)
    {
      const char*dylib_close = getenv("PUK_DYLIB_CLOSE");
      if(dylib_close == NULL)
	{
	  if(first)
	    {
	      padico_warning("Puk: using RTLD_NOCLOSE by default on valgrind to keep dynamic objects in memory. Define PUK_DYLIB_CLOSE=1 to disable this behavior.\n");
	      first = 0;
	    }
	  mode |= RTLD_NODELETE;
	}
    }
  puk_dylib_lock();
  void*p = dlopen(pathname, mode);
  if(p == NULL)
    {  
      int retry = 10;
      while(p == NULL && retry-- > 0)
	{
	  usleep(100000);
	  p = dlopen(pathname, mode);
	}
    }
  puk_dylib_unlock();
  return p;
}

int puk_dylib_close(void *handle)
{
  puk_dylib_lock();
  int rc = dlclose(handle);
  puk_dylib_unlock();
  return rc;
}

char*puk_dylib_error(void)
{
  puk_dylib_lock();
  char*rc = dlerror();
  puk_dylib_unlock();
  return rc;
}

void*puk_dylib_resolve(void*handle, const char*symbol) 
{
  void*rc;
  const int symbol_maxlen = 256;
  char mangled_symbol[symbol_maxlen];
  snprintf(mangled_symbol, symbol_maxlen, "%s%s", SYMBOL_PREFIX, symbol);
  puk_dylib_lock();
  rc = dlsym(handle, mangled_symbol);
  if(rc == NULL)
    {
      rc = dlsym(handle, symbol);
    }
  puk_dylib_unlock();
  return rc;
}


/* ********************************************************** */

padico_rc_t puk_lib_open(puk_lib_t*lib, const char*lib_file)
{
  padico_rc_t rc = NULL;
  *lib = puk_lib_new();
  (*lib)->filename = padico_strdup(lib_file);
  padico_out(20, "LIB LOAD loading %s\n", lib_file);
  
  if(access((*lib)->filename, R_OK) != 0)
    {
      rc = padico_rc_error("LIB load error: %s not found", lib_file);
      return rc;
    }
  
  (*lib)->dl_handle = puk_dylib_open((*lib)->filename, PUK_DYLIB_DEFAULT_MODE);
  if(!(*lib)->dl_handle) 
    {
      rc = padico_rc_error("LIB load error: puk_dylib_open() failed for %s (%s)", 
			   lib_file, puk_dylib_error());
      return rc;
    }
  padico_out(20, "LIB LOAD %s ok\n", lib_file);
  return rc;
}

void puk_lib_setself(puk_lib_t lib, puk_mod_t mod)
{
  typedef void (*setself_fun_t)(puk_mod_t);
  setself_fun_t setself;
  padico_out(40, "LIB INIT **%s**\n", lib->filename);
  setself = (setself_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_setself");
  if(setself != NULL)
    {
      (*setself)(mod);
    }
}

int puk_lib_init(puk_lib_t lib)
{
  puk_init_fun_t init_fun;
  int rc = 0;
  init_fun = (puk_init_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_init");
  if(init_fun != NULL)
    {
      rc = (*init_fun)();
    }
  return rc;
}

puk_run_fun_t puk_lib_run_resolve(puk_lib_t lib)
{
  puk_run_fun_t fun_ptr;

  padico_out(50, "LIB RUN try resolve in %s\n", lib->filename);
  if(lib == NULL)
    {
      return NULL;
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_run")) != NULL)
    {
      padico_out(40, "LIB RUN padico_module_run() found --PadicoTM binary code.\n");
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "marcel_main")) != NULL)
    {
      padico_out(40, "LIB RUN marcel_main() found --Marcel binary code.\n");
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "main")) != NULL)
    {
      padico_out(40, "LIB RUN main() found --standard binary code.\n");
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "MAIN__")) != NULL)
    {
      padico_out(40, "LIB RUN MAIN__() found --standard FORTRAN code.\n");
    }
  else
    {
      padico_out(6, "no runnable function found in lib %s.\n", lib->filename);
    }
  return fun_ptr;
}

padico_rc_t puk_lib_unload(puk_lib_t lib)
{
  puk_finalize_fun_t fun_ptr;
  padico_rc_t rc = NULL;
  int err = -1;
  
  if(lib == NULL)
    {
      rc = padico_rc_error("LIB UNLOAD: lib %s was not succesfully initialized."
			   " Cannot unload.\n", lib->filename);
      return rc;
    }
  fun_ptr = (puk_finalize_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_finalize");
  if (fun_ptr != NULL)
    {
      (*fun_ptr)();
    }
  padico_out(50, "LIB closing lib.\n");
  err = puk_dylib_close(lib->dl_handle);
  padico_out(50, "LIB UNLOAD %s (dlclose rc=%d)\n", lib->filename, err);
  if(err)
    {
      char*msg = puk_dylib_error();
      padico_warning("puk_dylib_close() error *%s*\n", msg);
      rc = padico_rc_error("LIB unload error: dlclose() failed for lib %s", lib->filename);
      return rc;
    }
  padico_out(50, "LIB UNLOAD %s ok\n",
	       lib->filename);
  puk_lib_delete(lib);
  return rc;
}


/***************************************************/

void puk_lib_ctor(puk_lib_t l)
{
  l->filename = NULL;
  l->dl_handle = NULL;
}

void puk_lib_dtor(puk_lib_t l) 
{
  if(l->filename)
    padico_free(l->filename);
}


