/** @file
 * @brief Provides "builtin" module driver
 * @ingroup Puk
 */

#include "Puk-internals.h"
#include "Module.h"

PADICO_MODULE_HOOK(Puk);

/******************************************************/

PUK_VECT_TYPE(puk_mod_builtin, const struct puk_mod_builtin_s*);

static struct
{
  puk_mod_builtin_vect_t builtins; /**< vector of builtins modules (already loaded in binary) */
  puk_component_t builtin_loader;
} puk_builtin = { .builtins = NULL };

/******************************************************/

static padico_rc_t puk_unit_builtin_load  (puk_unit_t unit);
static padico_rc_t puk_unit_builtin_start (puk_unit_t unit, puk_job_t job);
static padico_rc_t puk_unit_builtin_unload(puk_unit_t unit);

static const struct padico_loader_s puk_builtin_driver =
{
  .name   = "builtin",
  .load   = &puk_unit_builtin_load,
  .start  = &puk_unit_builtin_start,
  .unload = &puk_unit_builtin_unload,
};

void padico_puk_builtin_init(void)
{
  puk_builtin.builtin_loader =
    puk_component_declare("PadicoLoader-builtin", puk_component_provides("PadicoLoader", "loader", &puk_builtin_driver));
}

void padico_puk_builtin_finalize(void)
{
  /*
    puk_mod_builtin_vect_itor_t i;
    puk_vect_foreach(i, puk_mod_builtin, puk_builtin.builtins)
    {
    const struct puk_mod_builtin_s*const b = *i;
    }
  */
  puk_mod_builtin_vect_delete(puk_builtin.builtins);
}


/* *** 'builtin' toolbox *********************************** */

/* declares a builtin module- used internally by Module.h macros 
 * WARNING- called as static initializer before puk_init()
 */
void puk_mod_builtin_declare(const struct puk_mod_builtin_s*builtin)
{
  if(puk_builtin.builtins == NULL)
    puk_builtin.builtins = puk_mod_builtin_vect_new();
  puk_mod_builtin_vect_push_back(puk_builtin.builtins, builtin);
}

/** tries to load module as built-in.
 */
void puk_mod_builtin_tryload(puk_mod_t*mod, const char*mod_name)
{
  puk_lock_acquire();
  if(puk_builtin.builtins != NULL)
    {
      puk_mod_builtin_vect_itor_t i;
      puk_vect_foreach(i, puk_mod_builtin, puk_builtin.builtins)
	{
	  const struct puk_mod_builtin_s*const b = *i;
	  if(strcmp(b->name, mod_name) == 0)
	    {
	      *mod = puk_mod_new(mod_name, "builtin");
	      padico_string_t unit_name = padico_string_new();
	      padico_string_printf(unit_name, "built-in unit %s", mod_name);
	      struct puk_unit_s unit =
		{
		  .name            = padico_strdup(padico_string_get(unit_name)),
		  .mod             = *mod,
		  .driver_specific = (void*)b->funcs
		};
	      padico_string_delete(unit_name);
	      *(b->mod) = *mod;
	      puk_unit_vect_push_back(&(*mod)->units, unit);
	      goto out;
	    }
	}
    }
 out:
  puk_lock_release();
  return;
}


/* *** driver: "builtin" *********************************** */

static padico_rc_t puk_unit_builtin_load(puk_unit_t unit)
{
  padico_rc_t rc = NULL;
  const struct puk_mod_funcs_s*funcs = unit->driver_specific;
  if(funcs->init)
    {
      int err = (*funcs->init)();
      if(err)
        {
          padico_warning("Puk: built-in module %s init error.\n", unit->mod->mod_name);
          rc = padico_rc_error("built-in module *%s* load error- module initialization rc=%d",
                               unit->mod->mod_name, err);
        }
    }
  return rc;
}

static padico_rc_t puk_unit_builtin_start(puk_unit_t unit, puk_job_t job)
{
  padico_rc_t rc = NULL;
  const struct puk_mod_funcs_s*funcs = unit->driver_specific;
  if(funcs->run)
    {
      (*funcs->run)(0, NULL);
    }
  return rc;
}

static padico_rc_t puk_unit_builtin_unload(puk_unit_t unit)
{
  padico_rc_t rc = NULL;
  const struct puk_mod_funcs_s*funcs = unit->driver_specific;
  if(funcs->finalize)
    {
      (*funcs->finalize)();
    }
  unit->driver_specific = NULL;
  padico_free(unit->name);
  unit->name = NULL;
  return rc;
}


