/*
 * NewMadeleine
 * Copyright (C) 2011-2019 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <values.h>
 #include <ctype.h>

#include <nm_private.h>
#include <nm_minidriver.h>

#include <Padico/Module.h>
#include <Padico/Puk.h>
#include <Padico/AddrDB.h>
#include <Padico/Topology.h>

PADICO_MODULE_DECLARE(nm_driver_test);
PADICO_MODULE_INIT(nm_driver_test_init);
PADICO_MODULE_RUN(nm_driver_test_run);

/* hash minidriver recptacles by status (void*) */
PUK_HASHTABLE_TYPE(nm_minidriver, void*, struct puk_receptacle_NewMad_minidriver_s*,
                   &puk_hash_pointer_default_hash, &puk_hash_pointer_default_eq, NULL);

static int rank = -1, nprocs = -1;
static nm_len_t max_size = NM_LEN_UNDEFINED;

static struct nm_driver_test_s
{
  const struct nm_minidriver_iface_s*driver;
  struct nm_minidriver_properties_s props;
  puk_context_t context;
  struct puk_receptacle_NewMad_minidriver_s*trks;
  struct nm_minidriver_hashtable_s trks_hash;
  struct
  {
    struct iovec v;
    struct nm_data_s data;
  } send;
  struct
  {
    struct iovec v;
    struct nm_data_s data;
  } recv;
} nm_driver_test;

static inline void nm_driver_test_isend(struct puk_receptacle_NewMad_minidriver_s*r, char*buf, size_t len)
{
  if(r->driver->send_buf_get)
    {
      void*p_driver_buf = NULL;
      nm_len_t driver_len = NM_LEN_UNDEFINED;
      (*r->driver->send_buf_get)(r->_status, &p_driver_buf, &driver_len);
      memcpy(p_driver_buf, buf, len);
      (*r->driver->send_buf_post)(r->_status, len);
    }
  else if(r->driver->send_iov_post)
    {
      nm_driver_test.send.v = (struct iovec){ .iov_base = buf, .iov_len = len };
      (*r->driver->send_iov_post)(r->_status, &nm_driver_test.send.v, 1);
    }
  else
    {
      nm_data_contiguous_build(&nm_driver_test.send.data, buf, len);
      (*r->driver->send_data_post)(r->_status, &nm_driver_test.send.data, 0, len);
    }
}

static inline int nm_driver_test_spoll(struct puk_receptacle_NewMad_minidriver_s*r, char*buf, size_t len)
{
  return (*r->driver->send_poll)(r->_status);
}

static inline void nm_driver_test_send(struct puk_receptacle_NewMad_minidriver_s*r, char*buf, size_t len)
{
  nm_driver_test_isend(r, buf, len);
  int rc = -1;
  do
    {
      rc = nm_driver_test_spoll(r, buf, len);
    }
  while(rc != 0);
}

static inline void nm_driver_test_irecv(struct puk_receptacle_NewMad_minidriver_s*r, char*buf, size_t len)
{
  if(r->driver->recv_buf_poll)
    {
      /* nothing to do */
    }
  else
    {
      if(r->driver->recv_iov_post)
        {
          nm_driver_test.recv.v = (struct iovec){ .iov_base = buf, .iov_len = len };
          (*r->driver->recv_iov_post)(r->_status, &nm_driver_test.recv.v, 1);
        }
      else
        {
          nm_data_contiguous_build(&nm_driver_test.recv.data, buf, len);
          (*r->driver->recv_data_post)(r->_status, &nm_driver_test.recv.data, 0, len);
        }
    }
}

static inline int nm_driver_test_rpoll(struct puk_receptacle_NewMad_minidriver_s*r, char*buf, size_t len)
{
  if(r->driver->recv_buf_poll)
    {
      void*p_driver_buf = NULL;
      nm_len_t driver_len = NM_LEN_UNDEFINED;
      int rc = (*r->driver->recv_buf_poll)(r->_status, &p_driver_buf, &driver_len);
      if(rc == 0)
        {
          memcpy(buf, p_driver_buf, len);
          (*r->driver->recv_buf_release)(r->_status);
        }
      return rc;
    }
  else
    {
      return (*r->driver->recv_poll_one)(r->_status);
    }
}

static inline void nm_driver_test_recv(struct puk_receptacle_NewMad_minidriver_s*r, char*buf, size_t len)
{
  nm_driver_test_irecv(r, buf, len);
  int rc = -1;
  do
    {
      rc = nm_driver_test_rpoll(r, buf, len);
    }
  while(rc != 0);
}

static inline void nm_driver_test_recv_probe_any(void**_status)
{
  (*nm_driver_test.driver->recv_probe_any)(nm_driver_test.context, _status);
}

static inline int nm_driver_test_recv_cancel(struct puk_receptacle_NewMad_minidriver_s*r)
{
  return (*r->driver->recv_cancel)(r->_status);
}

#ifdef PIOMAN_MULTITHREAD
struct nm_driver_test_wait_s
{
  struct puk_receptacle_NewMad_minidriver_s*r;
  char*buf;
  size_t len;
};
static void*nm_driver_test_wait_any_worker(void*_arg)
{
  void*status = NULL;
  int rc = (*nm_driver_test.driver->recv_wait_any)(nm_driver_test.context, &status);
  if(rc == NM_ESUCCESS)
    {
      struct nm_driver_test_wait_s*p_wait = _arg;
      nm_driver_test_recv(p_wait->r, p_wait->buf, p_wait->len);
    }
  return NULL;
}
#endif /* PIOMAN_MULTITHREAD */

int nm_driver_test_init(void)
{
  const char*s_component = padico_getenv("NM_DRIVER_TEST");
  fprintf(stderr, "# nm_driver_test: using driver: NM_DRIVER_TEST = %s\n", s_component);
  assert(s_component != NULL);
  /* ** resolve driver */
  puk_component_t p_component0 = puk_component_resolve(s_component);
  puk_component_t component = NULL;
  if(p_component0 == NULL)
    {
      NM_FATAL("nm_driver_test: cannot resolve component '%s'\n", s_component);
    }
  puk_context_t p_context0 = puk_component_get_context(p_component0, puk_iface_NewMad_minidriver(), NULL);
  if(p_context0 == NULL)
    {
      fprintf(stderr, "# nm_driver_test: no context in component.\n");
      puk_component_t p_composite = puk_composite_new("nm_driver_test");
      puk_context_t p_context0 = puk_context_new(p_component0, p_composite, NULL);

      puk_composite_add_entrypoint(p_composite, puk_iface_NewMad_minidriver(), "minidriver",
                                   NULL /* provider_port */, p_context0->id);

      /* plug ibverbs component if needed */
      puk_iface_t ibverbs_iface = puk_iface_register("NewMad_ibverbs");
      puk_component_conn_t p_ibverbs_conn = puk_context_conn_lookup(p_context0, ibverbs_iface, NULL);
      if(p_ibverbs_conn != NULL)
        {
          fprintf(stderr, "# nm_driver_test: component depends on NewMad_ibverbs common.\n");
          puk_component_t p_ibverbs_component = puk_component_resolve("NewMad_ibverbs_common");
          puk_context_t p_ibverbs_context = puk_context_new(p_ibverbs_component, p_composite, NULL);
          puk_context_conn_connect(p_context0, NULL, p_ibverbs_context->id, ibverbs_iface, "ibverbs");
        }
      component = p_composite;
    }
  else
    {
      fprintf(stderr, "# nm_driver_test: component has a context.\n");
      component = puk_composite_duplicate(p_component0, "nm_driver_test");
    }
  assert(component != NULL);
  fprintf(stderr, "# nm_driver_test: component loaded.\n");
  padico_string_t s0 = puk_component_serialize_id(p_component0, "nm_driver_test");
  fprintf(stderr, "# nm_driver_test: driver component: \n%s\n", padico_string_get(s0));
  padico_string_delete(s0);
  padico_string_t s = puk_component_serialize_id(component, "nm_driver_test");
  fprintf(stderr, "# nm_driver_test: contextualized component: \n%s\n", padico_string_get(s));
  padico_string_delete(s);

  /* ** get component context and driver */
  nm_driver_test.context = puk_component_get_context(component, puk_iface_NewMad_minidriver(), NULL);
  if(!nm_driver_test.context)
    {
      NM_FATAL("nm_driver_test: cannot find context for Minidriver facet in component %s. Cannot instantiate.\n",
               component->name);
    }
  nm_driver_test.driver = puk_component_get_driver_NewMad_minidriver(component, NULL);

  /* ** get driver properties */
  nm_driver_test.props = (struct nm_minidriver_properties_s){ 0 };
#ifdef PIOMAN_TOPOLOGY_HWLOC
  nm_driver_test.props.profile.cpuset = hwloc_bitmap_alloc();
#endif
  fprintf(stderr, "# nm_driver_test: getting properties...\n");
  (*nm_driver_test.driver->getprops)(nm_driver_test.context, &nm_driver_test.props);
  const struct nm_minidriver_properties_s*props = &nm_driver_test.props;
  if(props->capabilities.max_msg_size > 0)
    {
      max_size = props->capabilities.max_msg_size;
      fprintf(stderr, "# nm_driver_test: max message size for driver %s = %lu\n", component->name, max_size);
    }
  fprintf(stderr, "#     max_msg_size = %lu; supports_data = %d; supports_buf_send = %d; supports_buf_recv = %d; supports_wait_any = %d\n",
          props->capabilities.max_msg_size, props->capabilities.supports_data,
          props->capabilities.supports_buf_send, props->capabilities.supports_buf_recv, props->capabilities.supports_wait_any);
  fprintf(stderr, "#     prefers_wait_any = %d; no_iovec = %d; has_recv_any = %d; min_period = %d; is_exportable = %d; trk_rdv = %d\n",
          props->capabilities.prefers_wait_any, props->capabilities.no_iovec, props->capabilities.has_recv_any,
          props->capabilities.min_period, props->capabilities.is_exportable, props->capabilities.trk_rdv);

  /* ** initialize driver & get url */
  fprintf(stderr, "# nm_driver_test: initializing driver...\n");
  const void*drv_url = NULL;
  size_t url_size = 0;
  (*nm_driver_test.driver->init)(nm_driver_test.context, &drv_url, &url_size);
  fprintf(stderr, "# nm_driver_test: init done; url_size = %lu\n", url_size);
  fprintf(stderr, "# nm_driver_test: local url = ");
  int i;
  for(i = 0; i < url_size; i++)
    {
      const char*u = drv_url;
      if(isprint(u[i]))
        fprintf(stderr, "%c ", u[i]);
      else
        fprintf(stderr, "[0x%x] ", u[i]);
    }
  fprintf(stderr, "\n");

  const char*session = padico_na_node_getsession(padico_na_getlocalnode());
  fprintf(stderr, "# nm_driver_test: connecting nodes in session %s\n", session);
  nprocs = padico_na_size();
  nm_driver_test.trks = malloc(nprocs * sizeof(struct puk_receptacle_NewMad_minidriver_s));
  nm_minidriver_hashtable_init(&nm_driver_test.trks_hash);
  int k;
  for(k = 0; k < nprocs; k++)
    {
      const padico_na_node_t n = padico_na_session_getnodebyrank(session, k);
      if(padico_na_getlocalnode() != n)
        {
          /* instantiate & resolve minidriver interface */
          puk_instance_t instance = puk_component_instantiate(component);
          puk_instance_indirect_NewMad_minidriver(instance, "minidriver", &nm_driver_test.trks[k]);
          nm_minidriver_hashtable_insert(&nm_driver_test.trks_hash, nm_driver_test.trks[k]._status, &nm_driver_test.trks[k]);
          /* ** connect */
          const char addr_name[] = "nm_driver_test";
          fprintf(stderr, "# nm_driver_test: connecting to node #%d: %s\n", k, padico_na_node_getname(n));
          padico_addrdb_publish(n, addr_name, NULL, 0, drv_url, url_size);
          const int peer_url_size = 64;
          char peer_url[peer_url_size];
          padico_req_t req = padico_tm_req_new(NULL, NULL);
          padico_addrdb_get(n, addr_name, NULL, 0, &peer_url[0], peer_url_size, req);
          padico_tm_req_wait(req);
          (*nm_driver_test.trks[k].driver->connect)(nm_driver_test.trks[k]._status, peer_url, url_size);
        }
      else
        {
          nm_driver_test.trks[k].driver = NULL;
          rank = k;
        }
    }
  
  return 0;
}

int nm_driver_test_run(int argc, char**argv)
{
  fprintf(stderr, "# nm_driver_test: running....\n");
  const int default_max_size = 128 * 1024;
  int buf_size = ((max_size > 0) && (max_size < default_max_size)) ? max_size : default_max_size;
  const int next = (rank + 1) % nprocs; 
  const int prev = (rank - 1 + nprocs) % nprocs;
  fprintf(stderr, "# nm_driver_test: rank = %d; prev = %d; next = %d\n", rank, prev, next);
  void*buf = malloc(buf_size);
  int i;
  
  /* blocking send/recv, all sizes */
  fprintf(stderr, "# ##\n");
  fprintf(stderr, "# nm_driver_test: testing blocking send/recv...\n");
  for(i = 1; i < buf_size; i++)
    {
      if((i == 1) || (i % 1024 == 0))
        {
          fprintf(stderr, "\n");
          fprintf(stderr, "# nm_driver_test: blocking send/recv; msg size = %d... ", i);
        }
      if(i % 32 == 0)
        {
          fprintf(stderr, ".");
          fflush(stderr);
        }
      if(rank % 2 == 0)
        {
          nm_driver_test_send(&nm_driver_test.trks[prev], buf, i); 
          nm_driver_test_recv(&nm_driver_test.trks[next], buf, i);
        }
      else
        {
          nm_driver_test_recv(&nm_driver_test.trks[next], buf, i);
          nm_driver_test_send(&nm_driver_test.trks[prev], buf, i); 
        }
    }
  
  /* non-blocking send/recv */
  fprintf(stderr, "# ##\n");
  fprintf(stderr, "# nm_driver_test: testing non-blocking send/recv...\n");
  for(i = 1; i < buf_size; i++)
    {
      if((i == 1) || (i % 1024 == 0))
        {
          fprintf(stderr, "\n");
          fprintf(stderr, "# nm_driver_test: non-blocking send/recv; msg size = %d... ", i);
        }
      if(i % 32 == 0)
        {
          fprintf(stderr, ".");
          fflush(stderr);
        }
      nm_driver_test_irecv(&nm_driver_test.trks[next], buf, i);
      nm_driver_test_isend(&nm_driver_test.trks[prev], buf, i);
      int rrc = -1, src = -1;
      do
        {
          if(rrc != 0)
            rrc = nm_driver_test_rpoll(&nm_driver_test.trks[next], buf, i);
          if(src != 0)
            src = nm_driver_test_spoll(&nm_driver_test.trks[prev], buf, i);
        }
      while((rrc != 0) || (src != 0));
    }

  /* recv any */
  fprintf(stderr, "# ##\n");
  fprintf(stderr, "# nm_driver_test: testing recv_probe_any...\n");
  if(nm_driver_test.props.capabilities.has_recv_any && nm_driver_test.driver->recv_probe_any)
    {
      for(i = 1; i < buf_size; i = i * 1.1 + 1)
        {
          const nm_len_t size = 1;
          fprintf(stderr, "# nm_driver_test: testing recv_probe_any; size = %d\n", i);
          nm_driver_test_isend(&nm_driver_test.trks[prev], buf, size);
          int rc = -1;
          void*status = NULL;
          do
            {
              if(status == NULL)
                nm_driver_test_recv_probe_any(&status);
              if(rc != 0)
                rc = nm_driver_test_spoll(&nm_driver_test.trks[prev], buf, size);
            }
          while((status == NULL) || (rc != 0));
          struct puk_receptacle_NewMad_minidriver_s*trk = nm_minidriver_hashtable_lookup(&nm_driver_test.trks_hash, status);
          assert(trk != NULL);
          nm_driver_test_recv(trk, buf, size);
          fprintf(stderr, "# nm_driver_test: recv_probe_any ok.\n");
        }
    }
  else
    {
      fprintf(stderr, "# nm_driver_test: recv_probe_any not available.\n");
    }

#ifdef PIOMAN_MULTITHREAD
  /* wait any*/
  fprintf(stderr, "# ##\n");
  fprintf(stderr, "# nm_driver_test: testing recv_wait_any...\n");
  if(nm_driver_test.props.capabilities.supports_wait_any && nm_driver_test.driver->recv_wait_any)
    {
      for(i = 1; i < buf_size; i = i * 1.1 + 1)
        {
          const nm_len_t size = 1;
          fprintf(stderr, "# nm_driver_test: testing recv_wait_any; size = %d\n", i);
          nm_driver_test_isend(&nm_driver_test.trks[prev], buf, size);
          struct nm_driver_test_wait_s wait =
            {
              .r = &nm_driver_test.trks[next],
              .buf = buf,
              .len = size
            };
          piom_thread_t tid;
          piom_thread_create(&tid, NULL, &nm_driver_test_wait_any_worker, &wait);
          nm_driver_test_send(&nm_driver_test.trks[prev], buf, size); 
          piom_thread_join(tid);
          fprintf(stderr, "# nm_driver_test: recv_wait_any ok.\n");
        }
    }
  else
    {
      fprintf(stderr, "# nm_driver_test: recv_wait_any not available.\n");
    }
#else /* PIOMAN_MULTITHREAD */
  fprintf(stderr, "# nm_driver_test: pioman disabled or non-threaded; not testing recv_wait_any.\n");
#endif /* PIOMAN_MULTITHREAD */

  /* cancel */
  fprintf(stderr, "# ##\n");
  fprintf(stderr, "# nm_driver_test: testing recv_cancel...\n");
  if(nm_driver_test.driver->recv_cancel)
    {
      const nm_len_t size = 4;
      nm_driver_test_irecv(&nm_driver_test.trks[next], buf, size);
      int rc = nm_driver_test_recv_cancel(&nm_driver_test.trks[next]);
      fprintf(stderr, "# nm_driver_test: after recv- recv_cancel rc = %d (%s)\n", rc, nm_strerror(rc));
      rc = nm_driver_test_recv_cancel(&nm_driver_test.trks[next]);
      fprintf(stderr, "# nm_driver_test: not after recv- recv_cancel rc = %d (%s)\n", rc, nm_strerror(rc));
    }
  else
    {
      fprintf(stderr, "# nm_driver_test: recv_cancel not available; not tested.\n");
    }
  
  return 0;
}

