/*
 * NewMadeleine
 * Copyright (C) 2014-2018 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <nm_public.h>
#include <nm_private.h>
#include <nm_sendrecv_interface.h>
#include <nm_launcher_interface.h>
#include <Padico/Puk.h>
#include "nm_coll_private.h"


/* ** ibcast *********************************************** */

/** status of a bcast */
struct nm_coll_bcast_s
{
  struct nm_coll_tree_status_s coll_tree;
  struct nm_data_s data;
};

static void nm_coll_ibcast_step(struct nm_coll_bcast_s*p_bcast);
static void nm_coll_ibcast_req_notifier(nm_sr_event_t event, const nm_sr_event_info_t*event_info, void*_ref);

static void nm_coll_ibcast_step(struct nm_coll_bcast_s*p_bcast)
{
 restart:
  assert(p_bcast->coll_tree.pending_reqs == 0);
  if(p_bcast->coll_tree.s < p_bcast->coll_tree.tree.height)
    {
      int parent = -1;
      int n_children = 0;
      nm_coll_tree_step(&p_bcast->coll_tree.tree, p_bcast->coll_tree.s, &parent,
                        p_bcast->coll_tree.children, &n_children);
      p_bcast->coll_tree.s++;
      assert(!((parent != -1) && (n_children > 0)));
      if(parent != -1)
        {
          nm_coll_tree_recv(&p_bcast->coll_tree, parent, &p_bcast->data,
                            &p_bcast->coll_tree.parent_request, &nm_coll_ibcast_req_notifier);
        }
      else if(n_children > 0)
        {
          int i;
          for(i = 0; i < n_children; i++)
            {
              nm_coll_tree_send(&p_bcast->coll_tree, p_bcast->coll_tree.children[i], &p_bcast->data,
                                &p_bcast->coll_tree.children_requests[i], &nm_coll_ibcast_req_notifier);
            }
        }
      else
        {
          assert((parent == -1) && (n_children == 0));
          goto restart;
        }
    }
  else
    {
      /* notify completion */
      nm_cond_signal(&p_bcast->coll_tree.status, 1);
    }
}

static void nm_coll_ibcast_req_notifier(nm_sr_event_t event, const nm_sr_event_info_t*event_info, void*_ref)
{
  struct nm_coll_tree_status_s*p_coll_tree = _ref;
  struct nm_coll_bcast_s*p_bcast = nm_container_of(p_coll_tree, struct nm_coll_bcast_s, coll_tree);
  nm_atomic_dec(&p_bcast->coll_tree.pending_reqs);
  if(p_bcast->coll_tree.pending_reqs == 0)
    {
      nm_coll_ibcast_step(p_bcast);
    }
}

void nm_coll_ibcast_wait(struct nm_coll_bcast_s*p_bcast)
{
  nm_coll_tree_status_wait(&p_bcast->coll_tree);
  free(p_bcast);
}

struct nm_coll_bcast_s*nm_coll_group_data_ibcast(nm_session_t p_session, nm_group_t p_group, int root, int self,
                                                 struct nm_data_s*p_data, nm_tag_t tag)
{
  struct nm_coll_bcast_s*p_bcast = malloc(sizeof(struct nm_coll_bcast_s));
  p_bcast->data = *p_data;
  nm_coll_tree_status_init(&p_bcast->coll_tree, p_session, p_group, root, self, tag);
  nm_coll_ibcast_step(p_bcast);
  return p_bcast;
}


/* ** bcast ************************************************ */

void nm_coll_group_data_bcast(nm_session_t p_session, nm_group_t p_group, int root, int self,
                              struct nm_data_s*p_data, nm_tag_t tag)
{
  const enum nm_coll_tree_kind_e kind = NM_COLL_TREE_BINOMIAL;
  assert(nm_group_get_gate(p_group, self) == nm_launcher_self_gate());
  const int size = nm_group_size(p_group);
  struct nm_coll_tree_info_s tree;
  nm_coll_tree_init(&tree, kind, size, self, root);
  nm_sr_request_t*requests = malloc(tree.max_arity * sizeof(nm_sr_request_t));
  int*children = malloc(sizeof(int) * tree.max_arity);
  int s;
  for(s = 0; s < tree.height; s++)
    {
      int parent = -1;
      int n_children = 0;
      nm_coll_tree_step(&tree, s, &parent, children, &n_children);
      assert((parent == -1) || (n_children == 0) || (children[0] == -1));
      if(parent != -1)
        {
          nm_gate_t p_parent_gate = nm_group_get_gate(p_group, parent);
          nm_sr_request_t request;
          nm_sr_irecv_data(p_session, p_parent_gate, tag, p_data, &request);
          nm_sr_rwait(p_session, &request);
        }
      if(n_children > 0)
        {
          int i;
          for(i = 0; i < n_children; i++)
            {
              assert(children[i] != -1);
              nm_gate_t p_gate = nm_group_get_gate(p_group, children[i]);
              nm_sr_isend_data(p_session, p_gate, tag, p_data, &requests[i]);
            }
          for(i = 0; i < n_children; i++)
            {
              nm_sr_swait(p_session, &requests[i]);
            }
        }
    }
  free(children);
  free(requests);
}

void nm_coll_group_bcast(nm_session_t p_session, nm_group_t p_group, int root, int self,
			 void*buffer, nm_len_t len, nm_tag_t tag)
{
  assert(nm_group_get_gate(p_group, self) == nm_launcher_self_gate());
  struct nm_data_s data;
  nm_data_contiguous_build(&data, (void*)buffer, len);
  nm_coll_group_data_bcast(p_session, p_group, root, self, &data, tag);
}

void nm_coll_data_bcast(nm_comm_t p_comm, int root, struct nm_data_s*p_data, nm_tag_t tag)
{
  nm_coll_group_data_bcast(nm_comm_get_session(p_comm), nm_comm_group(p_comm),
                           root, nm_comm_rank(p_comm), p_data, tag);
}

void nm_coll_bcast(nm_comm_t p_comm, int root, void*buffer, nm_len_t len, nm_tag_t tag)
{
  nm_coll_group_bcast(nm_comm_get_session(p_comm), nm_comm_group(p_comm),
                      root, nm_comm_rank(p_comm), buffer, len, tag);
}


