/*
 * NewMadeleine
 * Copyright (C) 2006-2014 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#ifndef NM_MPI_P2P_H
#define NM_MPI_P2P_H

/** \addtogroup mpi_interface */
/* @{ */

/** @name Functions: Point-to-point communications */
/* @{ */

/**
 * Performs a standard-mode, blocking send.
 * @param buffer initial address of send buffer
 * @param count number of elements in send buffer
 * @param datatype datatype of each send buffer element
 * @param dest rank of destination
 * @param tag message tag
 * @param comm communicator
 * @return MPI status
 */
int MPI_Send(const void*buffer,
             int count,
             MPI_Datatype datatype,
             int dest,
             int tag,
             MPI_Comm comm);


int MPI_Issend(const void*buf,
	       int count,
	       MPI_Datatype datatype,
	       int dest,
               int tag,
	       MPI_Comm comm,
	       MPI_Request *request);

/**
 * Posts a standard-mode, non blocking send.
 * @param buffer initial address of send buffer
 * @param count number of elements in send buffer
 * @param datatype datatype of each send buffer element
 * @param dest rank of destination
 * @param tag message tag
 * @param comm communicator
 * @param request pointer to request
 * @return MPI status
 */
int MPI_Isend(const void*buffer,
              int count,
              MPI_Datatype datatype,
              int dest,
              int tag,
              MPI_Comm comm,
              MPI_Request *request);

/**
 * Performs a ready-mode, blocking send.
 * @param buffer initial address of send buffer
 * @param count number of elements in send buffer
 * @param datatype datatype of each send buffer element
 * @param dest rank of destination
 * @param tag message tag
 * @param comm communicator
 * @return MPI status
 */
int MPI_Rsend(const void*buffer,
              int count,
              MPI_Datatype datatype,
              int dest,
              int tag,
              MPI_Comm comm);

/**
 * Performs a synchronous-mode, blocking send.
 * @param buffer initial address of send buffer
 * @param count number of elements in send buffer
 * @param datatype datatype of each send buffer element
 * @param dest rank of destination
 * @param tag message tag
 * @param comm communicator
 * @return MPI status
 */
int MPI_Ssend(const void* buffer,
              int count,
              MPI_Datatype datatype,
              int dest,
              int tag,
              MPI_Comm comm);

int MPI_Bsend(const void *buffer,
              int count,
              MPI_Datatype datatype,
              int dest,
              int tag,
              MPI_Comm comm);

int MPI_Irsend(const void*buf,
               int count,
               MPI_Datatype datatype,
               int dest,
               int tag,
               MPI_Comm comm,
               MPI_Request*request);
/**
 * Performs a standard-mode, blocking receive.
 * @param buffer initial address of receive buffer
 * @param count number of elements in receive buffer
 * @param datatype datatype of each receive buffer element
 * @param source rank of source
 * @param tag message tag
 * @param comm communicator
 * @param status status object
 * @return MPI status
 */
int MPI_Recv(void *buffer,
             int count,
             MPI_Datatype datatype,
             int source,
             int tag,
             MPI_Comm comm,
             MPI_Status *status);

/**
 * Posts a nonblocking receive.
 * @param buffer initial address of receive buffer
 * @param count number of elements in receive buffer
 * @param datatype datatype of each receive buffer element
 * @param source rank of source
 * @param tag message tag
 * @param comm communicator
 * @param request pointer to request
 * @return MPI status
 */
int MPI_Irecv(void* buffer,
              int count,
              MPI_Datatype datatype,
              int source,
              int tag,
              MPI_Comm comm,
              MPI_Request *request);


/**
 * Executes a blocking send and receive operation.
 * @param sendbuf initial address of send buffer
 * @param sendcount number of elements in send buffer
 * @param sendtype type of elements in send buffer
 * @param dest rank of destination
 * @param sendtag send tag
 * @param recvbuf initial address of receive buffer
 * @param recvcount number of elements in receive buffer
 * @param recvtype type of elements in receive buffer
 * @param source rank of source
 * @param recvtag receive tag
 * @param comm communicator
 * @param status status object
 * @return MPI status
 */
int MPI_Sendrecv(const void *sendbuf,
                 int sendcount,
                 MPI_Datatype sendtype,
                 int dest,
                 int sendtag,
                 void *recvbuf,
                 int recvcount,
                 MPI_Datatype recvtype,
                 int source,
                 int recvtag,
                 MPI_Comm comm,
                 MPI_Status *status);

int MPI_Sendrecv_replace(void*buf, int count, MPI_Datatype datatype, int dest, int sendtag, int source, int recvtag,
			 MPI_Comm comm, MPI_Status*status);



/**
 * Nonblocking operation that returns flag = true if there is a
 * message that can be received and that matches the message envelope
 * specified by source, tag and comm.
 * @param source source rank, or MPI_ANY_SOURCE
 * @param tag tag value or MPI_ANY_TAG
 * @param comm communicator
 * @param flag true if operation has completed
 * @param status status object
 * @return MPI status
 */
int MPI_Iprobe(int source,
               int tag,
               MPI_Comm comm,
               int *flag,
               MPI_Status *status);

/**
 * Blocks and returns only after a message that matches the message
 * envelope specified by source, tag and comm can be received.
 * @param source source rank, or MPI_ANY_SOURCE
 * @param tag tag value or MPI_ANY_TAG
 * @param comm communicator
 * @param status status object
 * @return MPI status
 */
int MPI_Probe(int source,
              int tag,
              MPI_Comm comm,
              MPI_Status *status);

/**
 * Computes the number of entries received.
 * @param status return status of receive operation
 * @param datatype datatype of each receive buffer entry
 * @param count number of received entries
 * @return MPI status
 */
int MPI_Get_count(MPI_Status *status,
                  MPI_Datatype datatype,
                  int *count);

int MPI_Get_elements(const MPI_Status*status,
		     MPI_Datatype datatype,
		     int *count);

int MPI_Get_elements_x(const MPI_Status*status,
		       MPI_Datatype datatype,
		       MPI_Count*count);

/* ********************************************************* */
/* PMPI aliases */

NM_MPI_PMPI(MPI_Issend);
NM_MPI_PMPI(MPI_Bsend);
NM_MPI_PMPI(MPI_Send);
NM_MPI_PMPI(MPI_Isend);
NM_MPI_PMPI(MPI_Irsend);
NM_MPI_PMPI(MPI_Rsend);
NM_MPI_PMPI(MPI_Ssend);
NM_MPI_PMPI(MPI_Recv);
NM_MPI_PMPI(MPI_Irecv);
NM_MPI_PMPI(MPI_Sendrecv);
NM_MPI_PMPI(MPI_Sendrecv_replace);
NM_MPI_PMPI(MPI_Iprobe);
NM_MPI_PMPI(MPI_Probe);

/* @}*/
/* @}*/

#endif /* NM_MPI_P2P_H */
