/*
 * NewMadeleine
 * Copyright (C) 2015-2018 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "nm_mpi_private.h"
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

#include <Padico/Module.h>
PADICO_MODULE_HOOK(MadMPI);

/** @name File */
/* @{ */

enum nm_mpi_file_op_e
  {
    NM_MPI_FILE_OP_NONE = 0,
    NM_MPI_FILE_OP_READ,
    NM_MPI_FILE_OP_WRITE
  };
struct nm_mpi_file_op_s
{
  enum nm_mpi_file_op_e op;
  struct nm_mpi_file_s*p_file;
  uint64_t offset;
  void*buf;
  int count;
  nm_mpi_datatype_t*p_datatype;
};

typedef struct nm_mpi_file_s
{
  MPI_File id; /**< identifier of the file */
  int amode;   /**< access mode */
  struct nm_mpi_communicator_s*p_comm;  /**< communicator attached to file */
  /** file view */
  struct
  {
    MPI_Offset disp;
    struct nm_mpi_datatype_s*p_etype;
    struct nm_mpi_datatype_s*p_filetype;
  } view;
  /** file kind */
  enum
    {
      NM_MPI_FILE_POSIX
    } kind;
  int fd;       /**< POSIX file descriptor */
  uint64_t cur; /**< current local pointer position */
  nm_mpi_errhandler_t*p_errhandler; /**< error handler attached to window */
  struct nm_mpi_file_op_s pending_op; /**< pending non-blocking operation */
} nm_mpi_file_t;

/* @} */

NM_MPI_HANDLE_TYPE(file, struct nm_mpi_file_s, _NM_MPI_FILE_OFFSET, 256);

static struct nm_mpi_handle_file_s nm_mpi_files;

/* ********************************************************* */

NM_MPI_ALIAS(MPI_File_open,               mpi_file_open);
NM_MPI_ALIAS(MPI_File_set_size,           mpi_file_set_size);
NM_MPI_ALIAS(MPI_File_get_size,           mpi_file_get_size);
NM_MPI_ALIAS(MPI_File_get_amode,          mpi_file_get_amode);
NM_MPI_ALIAS(MPI_File_get_group,          mpi_file_get_group);
NM_MPI_ALIAS(MPI_File_get_type_extent,    mpi_file_get_type_extent);
NM_MPI_ALIAS(MPI_File_set_view,           mpi_file_set_view);
NM_MPI_ALIAS(MPI_File_close,              mpi_file_close);
NM_MPI_ALIAS(MPI_File_delete,             mpi_file_delete);
NM_MPI_ALIAS(MPI_File_seek,               mpi_file_seek);
NM_MPI_ALIAS(MPI_File_create_errhandler,  mpi_file_create_errhandler);
NM_MPI_ALIAS(MPI_File_set_errhandler,     mpi_file_set_errhandler);
NM_MPI_ALIAS(MPI_File_get_errhandler,     mpi_file_get_errhandler);
NM_MPI_ALIAS(MPI_File_call_errhandler,    mpi_file_call_errhandler);
NM_MPI_ALIAS(MPI_File_sync,               mpi_file_sync);
NM_MPI_ALIAS(MPI_File_set_atomicity,      mpi_file_set_atomicity);

NM_MPI_ALIAS(MPI_File_read,               mpi_file_read);
NM_MPI_ALIAS(MPI_File_write,              mpi_file_write);
NM_MPI_ALIAS(MPI_File_read_all,           mpi_file_read);
NM_MPI_ALIAS(MPI_File_write_all,          mpi_file_write);
NM_MPI_ALIAS(MPI_File_read_at,            mpi_file_read_at);
NM_MPI_ALIAS(MPI_File_write_at,           mpi_file_write_at);
NM_MPI_ALIAS(MPI_File_read_at_all,        mpi_file_read_at_all);
NM_MPI_ALIAS(MPI_File_write_at_all,       mpi_file_write_at_all);

NM_MPI_ALIAS(MPI_File_read_all_begin,     mpi_file_read_all_begin);
NM_MPI_ALIAS(MPI_File_read_all_end,       mpi_file_read_all_end);
NM_MPI_ALIAS(MPI_File_write_all_begin,    mpi_file_write_all_begin);
NM_MPI_ALIAS(MPI_File_write_all_end,      mpi_file_write_all_end);

NM_MPI_ALIAS(MPI_File_read_at_all_begin,  mpi_file_read_at_all_begin);
NM_MPI_ALIAS(MPI_File_read_at_all_end,    mpi_file_read_at_all_end);
NM_MPI_ALIAS(MPI_File_write_at_all_begin, mpi_file_write_at_all_begin);
NM_MPI_ALIAS(MPI_File_write_at_all_end,   mpi_file_write_at_all_end);

/* ********************************************************* */

__PUK_SYM_INTERNAL
void nm_mpi_io_init(void)
{
  nm_mpi_handle_file_init(&nm_mpi_files);
}

__PUK_SYM_INTERNAL
void nm_mpi_io_exit(void)
{
  nm_mpi_handle_file_finalize(&nm_mpi_files, NULL /* &nm_mpi_files_free */);
}

__PUK_SYM_INTERNAL
nm_mpi_file_t*nm_mpi_file_get(MPI_File file)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, file);
  return p_file;
}

/* ********************************************************* */

int mpi_file_open(MPI_Comm comm, char*filename, int amode, MPI_Info info, MPI_File*fh)
{

  fprintf(stderr, "# madmpi: MPI_File_open = %s\n", filename);

  nm_mpi_file_t*p_file = nm_mpi_handle_file_alloc(&nm_mpi_files);
  nm_mpi_communicator_t*p_comm = nm_mpi_communicator_get(comm);
  if(p_comm == NULL)
    return MPI_ERR_COMM;
  if(nm_comm_rank(p_comm->p_nm_comm) < 0)
    return MPI_SUCCESS;
  p_file->amode = amode;
  p_file->p_comm = p_comm;
  p_file->kind = NM_MPI_FILE_POSIX;
  p_file->cur = 0;
  p_file->view.disp = 0;
  p_file->view.p_etype = NULL;
  p_file->view.p_filetype = NULL;
  const int mode =
    ((amode & MPI_MODE_RDONLY) ? O_RDONLY : 0) |
    ((amode & MPI_MODE_WRONLY) ? O_WRONLY : 0) |
    ((amode & MPI_MODE_RDWR)   ? O_RDWR   : 0) |
    ((amode & MPI_MODE_CREATE) ? O_CREAT  : 0) |
    ((amode & MPI_MODE_APPEND) ? O_APPEND : 0) ;
  int rc = -1;
  if(amode & MPI_MODE_EXCL)
    {
      if(nm_comm_rank(p_comm->p_nm_comm) == 0)
        {
          rc = NM_SYS(open)(filename, mode | O_EXCL, 0666);
        }
      nm_coll_bcast(p_comm->p_nm_comm, 0, &rc, sizeof(rc), NM_MPI_TAG_PRIVATE_FILE_OPEN);
      if(rc < 0)
        {
          return MPI_ERR_FILE;
        }
      if(nm_comm_rank(p_comm->p_nm_comm) != 0)
        {
          rc = NM_SYS(open)(filename, mode, 0666);
        }
    }
  else
    {
      rc = NM_SYS(open)(filename, mode, 0666);
    }
  if(rc >= 0)
    {
      p_file->fd = rc;
    }
  else
    {
      fprintf(stderr, "# open failed- rc = %d (%s).\n", rc, strerror(errno));
      return MPI_ERR_FILE;
    }
  *fh = p_file->id;
  return MPI_SUCCESS;
}

int mpi_file_close(MPI_File*fh)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, *fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  NM_SYS(close)(p_file->fd);
  nm_mpi_handle_file_free(&nm_mpi_files, p_file);
  *fh = MPI_FILE_NULL;
  return MPI_SUCCESS;
}

int mpi_file_delete(char*filename, MPI_Info info)
{
  int rc = NM_SYS(unlink)(filename);
  if(rc != 0)
    return MPI_ERR_NO_SUCH_FILE;
  return MPI_SUCCESS;
}

int mpi_file_sync(MPI_File fh)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  int rc = syncfs(p_file->fd);
  if(rc != 0)
    return MPI_ERR_IO;
  return MPI_SUCCESS;
}

int mpi_file_set_atomicity(MPI_File fh, int flag)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  if(flag != 0)
    {
      NM_MPI_WARNING("MPI_File_set_atomicity()- flag = 1. Not supported.\n");
      return MPI_ERR_OTHER;
    }
  return MPI_SUCCESS;
}

int mpi_file_seek(MPI_File fh, MPI_Offset offset, int whence)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  const int lwhence =
    ((whence == MPI_SEEK_SET) ? SEEK_SET : 0) |
    ((whence == MPI_SEEK_CUR) ? SEEK_CUR : 0) |
    ((whence == MPI_SEEK_END) ? SEEK_END : 0);
  if(lwhence == 0)
    return MPI_ERR_IO;
  off_t rc = NM_SYS(lseek)(p_file->fd, offset, lwhence);
  if(rc < 0)
    return MPI_ERR_IO;
  p_file->cur = rc;
  return MPI_SUCCESS;
}

int mpi_file_set_size(MPI_File fh, MPI_Offset size)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  int rc = NM_SYS(ftruncate)(p_file->fd, size);
  if(rc != 0)
    return MPI_ERR_IO;
  return MPI_SUCCESS;
}

int mpi_file_get_size(MPI_File fh, MPI_Offset*size)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  struct stat stat;
  int rc = fstat(p_file->fd, &stat);
  if(rc != 0)
    return MPI_ERR_IO;
  *size = stat.st_size;
  return MPI_SUCCESS;
}

int mpi_file_get_amode(MPI_File fh, int*amode)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  *amode = p_file->amode;
  return MPI_SUCCESS;
}

int mpi_file_get_group(MPI_File fh, MPI_Group*group)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  return mpi_comm_group(p_file->p_comm->id, group);
}

int mpi_file_get_type_extent(MPI_File fh, MPI_Datatype datatype, MPI_Aint*extent)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  if(p_datatype == NULL)
    return MPI_ERR_TYPE;
  if(extent != NULL)
    *extent = p_datatype->extent;
  return MPI_SUCCESS;
}

int mpi_file_set_view(MPI_File fh, MPI_Offset disp, MPI_Datatype etype, MPI_Datatype filetype, const char*datarep, MPI_Info info)
{
  if((datarep != NULL) && (strcmp(datarep, "native") != 0))
    return MPI_ERR_UNSUPPORTED_DATAREP;
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  nm_mpi_datatype_t*p_etype = nm_mpi_datatype_get(etype);
  if(p_etype == NULL)
    return MPI_ERR_TYPE;
  nm_mpi_datatype_t*p_filetype = nm_mpi_datatype_get(filetype);
  if(p_filetype == NULL)
    return MPI_ERR_TYPE;
  p_file->view.disp = disp;
  p_file->view.p_etype = p_etype;
  p_file->view.p_filetype = p_filetype;
  NM_SYS(lseek)(p_file->fd, disp, SEEK_SET);
  p_file->cur = disp;
  return MPI_SUCCESS;
}

/* ** File I/O ********************************************* */

struct nm_mpi_file_op_apply_s
{
  enum nm_mpi_file_op_e op;
  struct nm_mpi_file_s*p_file;
  void*buf;
  nm_len_t done;
};

static void nm_mpi_file_op_apply(void*ptr, nm_len_t len, void*_context)
{
  struct nm_mpi_file_op_apply_s*p_apply = _context;
  if(len > 0)
    {
      if(p_apply->op == NM_MPI_FILE_OP_READ)
        {
          const ssize_t rc = NM_SYS(pread)(p_apply->p_file->fd, p_apply->buf + p_apply->done, len, (uintptr_t)ptr);
          if(rc != len)
            {
              NM_MPI_FATAL_ERROR("read error rc = %lld; chunk size = %lld\n", (long long)rc, (long long)len);
            }
        }
      else if(p_apply->op == NM_MPI_FILE_OP_WRITE)
        {
          const ssize_t rc = NM_SYS(pwrite)(p_apply->p_file->fd, p_apply->buf + p_apply->done, len, (uintptr_t)ptr);
          if(rc != len)
            {
              NM_MPI_FATAL_ERROR("write error rc = %lld; chunk size = %lld\n", (long long)rc, (long long)len);
            }
        }
      else
        {
          NM_MPI_FATAL_ERROR("wrong op = %d in file op iterator.\n", p_apply->op);
        }
      p_apply->done += len;
    }
}

static void nm_mpi_file_op(nm_mpi_file_t*p_file, enum nm_mpi_file_op_e op, uint64_t offset, void*buf, int count,
                           nm_mpi_datatype_t*p_datatype)
{
  const size_t size = count * nm_mpi_datatype_size(p_datatype);
  void*ptr = malloc(size);
  if(op == NM_MPI_FILE_OP_WRITE)
    {
      nm_mpi_datatype_pack(ptr, buf, p_datatype, count);
    }
  struct nm_data_s disk_data;
  if(p_file->view.p_filetype != NULL)
    {
      const int unrolls = size / nm_mpi_datatype_size(p_file->view.p_filetype);
      nm_mpi_data_build(&disk_data, (void*)(uintptr_t)p_file->view.disp, p_file->view.p_filetype, unrolls);
    }
  else
    {
      nm_data_contiguous_build(&disk_data, NULL, offset + size);
    }
  struct nm_mpi_file_op_apply_s op_context =
    { .op = op, .p_file = p_file, .buf = ptr, .done = 0 };
  nm_data_chunk_extractor_traversal(&disk_data, offset, size, &nm_mpi_file_op_apply, &op_context);
  if(op == NM_MPI_FILE_OP_READ)
    {
      nm_mpi_datatype_unpack(ptr, buf, p_datatype, count);
    }
  free(ptr);
  p_file->cur = offset + size;
}

int mpi_file_read(MPI_File fh, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  nm_mpi_file_op(p_file, NM_MPI_FILE_OP_READ, p_file->cur, buf, count, p_datatype);
  if(status)
  {
    status->MPI_ERROR = MPI_SUCCESS;
    status->size = count * nm_mpi_datatype_size(p_datatype);
  }
  return MPI_SUCCESS;
}

int mpi_file_write(MPI_File fh, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  nm_mpi_file_op(p_file, NM_MPI_FILE_OP_WRITE, p_file->cur, buf, count, p_datatype);
  if(status)
  {
    status->MPI_ERROR = MPI_SUCCESS;
    status->size = count * nm_mpi_datatype_size(p_datatype);
  }
  return MPI_SUCCESS;
}

int mpi_file_read_at(MPI_File fh, MPI_Offset offset, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  nm_mpi_file_op(p_file, NM_MPI_FILE_OP_READ, offset, buf, count, p_datatype);
  if(status)
  {
    status->MPI_ERROR = MPI_SUCCESS;
    status->size = count * nm_mpi_datatype_size(p_datatype);
  }
  return MPI_SUCCESS;
}

int mpi_file_write_at(MPI_File fh, MPI_Offset offset, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  nm_mpi_file_op(p_file, NM_MPI_FILE_OP_WRITE, offset, buf, count, p_datatype);
  if(status)
  {
    status->MPI_ERROR = MPI_SUCCESS;
    status->size = count * nm_mpi_datatype_size(p_datatype);
  }
  return MPI_SUCCESS;
}

int mpi_file_read_all(MPI_File fh, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  return mpi_file_read(fh, buf, count, datatype, status);
}

int mpi_file_write_all(MPI_File fh, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  return mpi_file_write(fh, buf, count, datatype, status);
}

int mpi_file_read_at_all(MPI_File fh, MPI_Offset offset, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  return mpi_file_read_at(fh, offset, buf, count, datatype, status);
}

int mpi_file_write_at_all(MPI_File fh, MPI_Offset offset, void*buf, int count, MPI_Datatype datatype, MPI_Status*status)
{
  return mpi_file_write_at(fh, offset, buf, count, datatype, status);
}

static int nm_mpi_file_op_begin(struct nm_mpi_file_op_s*p_op, MPI_File fh, MPI_Offset offset, void*buf, int count, MPI_Datatype datatype)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  if(p_datatype == NULL)
    return MPI_ERR_TYPE;
  if(count < 0)
    return MPI_ERR_COUNT;
  p_op->op         = NM_MPI_FILE_OP_NONE;
  p_op->p_file     = p_file;
  p_op->offset     = offset;
  p_op->buf        = buf;
  p_op->count      = count;
  p_op->p_datatype = p_datatype;
  return MPI_SUCCESS;
}

static int nm_mpi_file_op_end(struct nm_mpi_file_op_s*p_op, MPI_Status*status)
{
  nm_mpi_file_op(p_op->p_file, p_op->op, p_op->offset, p_op->buf, p_op->count, p_op->p_datatype);
  if(status)
  {
    status->MPI_ERROR = MPI_SUCCESS;
    status->size = p_op->count * nm_mpi_datatype_size(p_op->p_datatype);
  }
  return MPI_SUCCESS;
}

int mpi_file_read_all_begin(MPI_File fh, void*buf, int count, MPI_Datatype datatype)
{
  struct nm_mpi_file_op_s op;
  int rc = nm_mpi_file_op_begin(&op, fh, 0, buf, count, datatype);
  if(!rc)
    {
      op.op = NM_MPI_FILE_OP_READ;
      op.p_file->pending_op = op;
    }
  return rc;
}
int mpi_file_read_all_end(MPI_File fh, void*buf, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  return nm_mpi_file_op_end(&p_file->pending_op, status);
}
int mpi_file_write_all_begin(MPI_File fh, const void*buf, int count, MPI_Datatype datatype)
{
  struct nm_mpi_file_op_s op;
  int rc = nm_mpi_file_op_begin(&op, fh, 0, (void*)buf, count, datatype);
  if(!rc)
    {
      op.op = NM_MPI_FILE_OP_WRITE;
      op.p_file->pending_op = op;
    }
  return rc;
}
int mpi_file_write_all_end(MPI_File fh, const void*buf, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  return nm_mpi_file_op_end(&p_file->pending_op, status);
}

int mpi_file_read_at_all_begin(MPI_File fh, MPI_Offset offset, void*buf, int count, MPI_Datatype datatype)
{
  struct nm_mpi_file_op_s op;
  int rc = nm_mpi_file_op_begin(&op, fh, offset, buf, count, datatype);
  if(!rc)
    {
      op.op = NM_MPI_FILE_OP_READ;
      op.p_file->pending_op = op;
    }
  return rc;
}
int mpi_file_read_at_all_end(MPI_File fh, void*buf, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  return nm_mpi_file_op_end(&p_file->pending_op, status);
}
int mpi_file_write_at_all_begin(MPI_File fh, MPI_Offset offset, const void*buf, int count, MPI_Datatype datatype)
{
  struct nm_mpi_file_op_s op;
  int rc = nm_mpi_file_op_begin(&op, fh, offset, (void*)buf, count, datatype);
  if(!rc)
    {
      op.op = NM_MPI_FILE_OP_WRITE;
      op.p_file->pending_op = op;
    }
  return rc;
}
int mpi_file_write_at_all_end(MPI_File fh, const void*buf, MPI_Status*status)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, fh);
  if(p_file == NULL)
    return MPI_ERR_BAD_FILE;
  return nm_mpi_file_op_end(&p_file->pending_op, status);
}


/* ** errhandler ******************************************* */

int mpi_file_create_errhandler(MPI_File_errhandler_fn*function, MPI_Errhandler*errhandler)
{
  return MPI_Errhandler_create(function, errhandler);
}

int mpi_file_set_errhandler(MPI_File file, MPI_Errhandler errhandler)
{
  struct nm_mpi_errhandler_s*p_errhandler = nm_mpi_errhandler_get(errhandler);
  if(p_errhandler == NULL)
    return MPI_ERR_ARG;
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, file);
  if(p_file == NULL)
    return MPI_ERR_FILE;
  p_file->p_errhandler = p_errhandler;
  return MPI_SUCCESS;
}

int mpi_file_get_errhandler(MPI_File file, MPI_Errhandler*errhandler)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, file);
  if(p_file == NULL)
    return MPI_ERR_FILE;
  struct nm_mpi_errhandler_s*p_new_errhandler = nm_mpi_errhandler_alloc();
  *p_new_errhandler = *p_file->p_errhandler;
  *errhandler = p_new_errhandler->id;
  return MPI_SUCCESS;
}

int mpi_file_call_errhandler(MPI_File file, int errorcode)
{
  nm_mpi_file_t*p_file = nm_mpi_handle_file_get(&nm_mpi_files, file);
  if(p_file == NULL)
    return MPI_ERR_FILE;
  p_file->p_errhandler->function(&file, &errorcode);
  return MPI_SUCCESS;
}
