/*
 * NewMadeleine
 * Copyright (C) 2006-2019 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULA R PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include <stdint.h>
#include <sys/uio.h>
#include <assert.h>

#include <nm_private.h>

PADICO_MODULE_HOOK(NewMad_Core);

static void nm_pkt_data_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s**pp_unpack,
				const struct nm_header_pkt_data_s*h, struct nm_pkt_wrap_s *p_pw);
static void nm_short_data_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s**pp_unpack,
				  const struct nm_header_short_data_s*h, struct nm_pkt_wrap_s *p_pw);
static void nm_small_data_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s**pp_unpack,
				  const struct nm_header_data_s*h, struct nm_pkt_wrap_s *p_pw);
static void nm_rdv_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s*p_unpack,
			   const struct nm_header_ctrl_rdv_s*h, struct nm_pkt_wrap_s *p_pw);

static void nm_core_unpack_unexpected(struct nm_core*p_core, struct nm_req_s**pp_unpack, struct nm_unexpected_s*p_unexpected);

/* ********************************************************* */
/* ** unexpected/matching */


/** fast allocator for struct nm_unexpected_s */
PUK_ALLOCATOR_TYPE(nm_unexpected, struct nm_unexpected_s);
static nm_unexpected_allocator_t nm_unexpected_allocator = NULL;
#define NM_UNEXPECTED_PREALLOC 256

static inline struct nm_unexpected_s*nm_unexpected_alloc(void)
{
  if(nm_unexpected_allocator == NULL)
    {
      nm_unexpected_allocator = nm_unexpected_allocator_new(NM_UNEXPECTED_PREALLOC);
    }
  struct nm_unexpected_s*p_unexpected = nm_unexpected_malloc(nm_unexpected_allocator);
  nm_unexpected_wildcard_list_cell_init(p_unexpected);
  nm_unexpected_gtag_list_cell_init(p_unexpected);
  nm_unexpected_gate_list_cell_init(p_unexpected);
  nm_unexpected_tag_list_cell_init(p_unexpected);
  return p_unexpected;
}

void nm_unexpected_wildcard_clean(struct nm_unexpected_wildcard_list_s*p_unexpected)
{
  struct nm_unexpected_s*p_chunk = nm_unexpected_wildcard_list_pop_front(p_unexpected);
  while(p_chunk)
    {
#ifdef DEBUG
      NM_DISPF("nmad: WARNING- chunk %p is still in use (gate = %p; seq = %d; tag = %llx:%llx)\n",
	      p_chunk, p_chunk->p_gate, p_chunk->seq,
	      (unsigned long long)nm_core_tag_get_hashcode(p_chunk->tag), (unsigned long long)nm_core_tag_get_tag(p_chunk->tag));
#endif /* DEBUG */
      if(p_chunk->p_pw)
	{
#if(defined(PIOMAN))
	  piom_ltask_completed(&p_chunk->p_pw->ltask);
#else /* PIOMAN */
	  nm_pw_ref_dec(p_chunk->p_pw);
#endif /* PIOMAN */
	}
      nm_unexpected_free(nm_unexpected_allocator, p_chunk);
      p_chunk = nm_unexpected_wildcard_list_pop_front(p_unexpected);
    }
}

void nm_unexpected_clean(struct nm_core*p_core)
{
  nm_matching_wildcard_table_destroy(&p_core->wildcard_table);
  nm_matching_tag_table_destroy(&p_core->tag_table);
  if(nm_unexpected_allocator != NULL)
    {
      nm_unexpected_allocator_delete(nm_unexpected_allocator);
      nm_unexpected_allocator = NULL;
    }
}

/** checks whether an unpack request matches a chunk from gate/seq/tag */
static inline int nm_unpack_is_matching(const struct nm_req_s*p_req, nm_gate_t p_gate, nm_seq_t seq, nm_core_tag_t tag, nm_seq_t next_seq)
{
  assert(p_gate != NULL);
  assert(!nm_status_test(p_req, NM_STATUS_FINALIZED));
  return (((p_req->p_gate == p_gate) || (p_req->p_gate == NM_ANY_GATE)) && /* gate matches */
	  nm_core_tag_match(tag, p_req->tag, p_req->unpack.tag_mask) && /* tag matches */
	  ((p_req->seq == seq) || ((p_req->seq == NM_SEQ_NONE) && (seq == next_seq))) /* seq number matches */ );
}

/** Find an unexpected chunk that matches an unpack request [p_gate, seq, tag]
 *  including matching with any_gate / any_tag in the unpack
 */
static struct nm_unexpected_s*nm_unexpected_find_matching(struct nm_core*p_core, struct nm_req_s*p_unpack)
{
  struct nm_unexpected_s*p_chunk = NULL;
  struct nm_gtag_s*p_gtag = p_unpack->p_gtag;
  assert(!nm_status_test(p_unpack, NM_STATUS_FINALIZED));
  assert(p_unpack->flags & NM_REQ_FLAG_UNPACK_MATCHED);
  nm_core_lock_assert(p_core);

  if(p_unpack->flags & NM_REQ_FLAG_MATCHING_FULL)
    {
      /* fully-specified requests, list per gtag */
      assert(p_gtag != NULL);
      const nm_seq_t next_seq = nm_seq_next(p_gtag->recv_seq_number);
      p_chunk = nm_unexpected_gtag_list_begin(&p_gtag->unexpected);
      if(p_chunk && nm_unpack_is_matching(p_unpack, p_chunk->p_gate, p_chunk->seq, p_chunk->tag, next_seq))
        {
          goto match_found;
        }
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_TAG)
    {
      /* only tag is specified, list per tag */
      struct nm_matching_tag_s*p_matching_tag = p_unpack->unpack.matching.p_matching_tag;
      p_chunk = nm_unexpected_tag_list_begin(&p_matching_tag->unexpected);
      if(p_chunk)
        {
          const nm_seq_t next_seq = nm_seq_next(p_chunk->p_gtag->recv_seq_number);
          if(nm_unpack_is_matching(p_unpack, p_chunk->p_gate, p_chunk->seq, p_chunk->tag, next_seq))
            {
              p_gtag = p_chunk->p_gtag;
              goto match_found;
            }
        }
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_GATE)
    {
      /* only gate is specified, list per gsession */
      struct nm_matching_gsession_s*p_matching_gate = p_unpack->unpack.matching.p_matching_gate;
      p_chunk = nm_unexpected_gate_list_begin(&p_matching_gate->unexpected);
      if(p_chunk)
        {
          const nm_seq_t next_seq = nm_seq_next(p_chunk->p_gtag->recv_seq_number);
          if(nm_unpack_is_matching(p_unpack, p_chunk->p_gate, p_chunk->seq, p_chunk->tag, next_seq))
            {
              p_gtag = p_chunk->p_gtag;
              goto match_found;
            }
        }
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_WILDCARD)
    {
      /* wildcard request or sparse bitmask for tag: full list lookup.
       * note: for wildcard requests, the first chunk will match. Linear lookup
       * only happens with sparse mask_tag.
       */
      struct nm_matching_wildcard_s*p_wildcard = p_unpack->unpack.matching.p_wildcard;
      puk_list_foreach(nm_unexpected_wildcard, p_chunk, &p_wildcard->unexpected)
	{
	  const nm_seq_t next_seq = nm_seq_next(p_chunk->p_gtag->recv_seq_number);
          if(nm_unpack_is_matching(p_unpack, p_chunk->p_gate, p_chunk->seq, p_chunk->tag, next_seq))
	    {
              p_gtag = p_chunk->p_gtag;
	      goto match_found;
	    }
	}
    }
  else
    {
      NM_FATAL("unspecified matching method.\n");
    }
  return NULL;

 match_found:

  if(p_unpack->seq == NM_SEQ_NONE)
    {
      p_gtag->recv_seq_number = nm_seq_next(p_gtag->recv_seq_number);
    }
  p_unpack->tag      = p_chunk->tag;
  p_unpack->unpack.tag_mask = NM_CORE_TAG_MASK_FULL;
  p_unpack->p_gate   = p_chunk->p_gate;
  p_unpack->seq      = p_chunk->seq;
  p_unpack->p_gtag   = p_gtag;
  return p_chunk;
}

/** find a matching request in the given list with gate/seq/tag */
static inline void nm_unpack_match_list(struct nm_req_s**pp_unpack_best, struct nm_req_list_s*p_list,
                                        const nm_gate_t p_gate, const nm_seq_t seq, const nm_seq_t next_seq, const nm_core_tag_t tag)
{
  struct nm_req_s*p_unpack_itor = NULL;
  /* we need a loop for out-of-order packets; anyway, usually the first req will match and early exit will happen */
  puk_list_foreach(nm_req, p_unpack_itor, p_list)
    {
      if(nm_unpack_is_matching(p_unpack_itor, p_gate, seq, tag, next_seq) &&
         ((*pp_unpack_best == NULL) || (p_unpack_itor->unpack.req_seq < (*pp_unpack_best)->unpack.req_seq)))
        {
          *pp_unpack_best = p_unpack_itor;
          return;
        }
      if((*pp_unpack_best) && (p_unpack_itor->unpack.req_seq > (*pp_unpack_best)->unpack.req_seq))
        {
          return;
        }
      if((p_unpack_itor->seq != NM_SEQ_NONE) && (p_unpack_itor->seq > seq))
        {
          return;
        }
    }
}

/** Find an unpack that matches a given packet that arrived from [p_gate, seq, tag]
 *  including matching with any_gate / any_tag in the unpack
 */
static struct nm_req_s*nm_unpack_find_matching(struct nm_core*p_core, nm_gate_t p_gate, nm_seq_t seq, nm_core_tag_t tag)
{
  assert(p_gate != NULL);
  struct nm_gtag_s*p_gtag = nm_gtag_get(&p_gate->tags, tag);
  const nm_seq_t next_seq = nm_seq_next(p_gtag->recv_seq_number);
  nm_core_lock_assert(p_core);
  struct nm_req_s*p_unpack = NULL;
  nm_unpack_match_list(&p_unpack, &p_gtag->unpacks, p_gate, seq, next_seq, tag);
  struct nm_matching_tag_s*p_matching_tag = nm_matching_tag_get(&p_core->tag_table, tag);
  nm_unpack_match_list(&p_unpack, &p_matching_tag->unpacks, p_gate, seq, next_seq, tag);
  struct nm_matching_gsession_s*p_matching_gsession = nm_matching_gsession_bytag(p_gate, tag);
  nm_unpack_match_list(&p_unpack, &p_matching_gsession->unpacks, p_gate, seq, next_seq, tag);
  struct nm_matching_wildcard_s*p_wildcard = nm_matching_wildcard_bytag(p_core, tag);
  nm_unpack_match_list(&p_unpack, &p_wildcard->unpacks, p_gate, seq, next_seq, tag);
  if(p_unpack)
    {
      if(p_unpack->seq == NM_SEQ_NONE)
	{
	  p_gtag->recv_seq_number = next_seq;
	}
      p_unpack->tag      = tag;
      p_unpack->unpack.tag_mask = NM_CORE_TAG_MASK_FULL;
      p_unpack->p_gate   = p_gate;
      p_unpack->seq      = seq;
      p_unpack->p_gtag   = p_gtag;
      return p_unpack;
    }
  return NULL;
}


/** mark 'chunk_len' data as received in the given unpack, and check for unpack completion
 * sets pp_unpack to NULL if finalized (req may have been destroyed by user handler) */
static inline void nm_core_unpack_check_completion(struct nm_core*p_core, struct nm_pkt_wrap_s*p_pw, struct nm_req_s**pp_unpack, nm_len_t chunk_len)
{
  struct nm_req_s*p_unpack = *pp_unpack;
  p_unpack->unpack.cumulated_len += chunk_len;
  if(p_unpack->unpack.cumulated_len == p_unpack->unpack.expected_len)
    {
      nm_core_lock_assert(p_core);
      struct nm_gtag_s*p_gtag = p_unpack->p_gtag;
      if(p_unpack->flags & NM_REQ_FLAG_MATCHING_WILDCARD)
	{
	  nm_req_list_remove(&p_unpack->unpack.matching.p_wildcard->unpacks, p_unpack);
	}
      else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_TAG)
        {
	  nm_req_list_remove(&p_unpack->unpack.matching.p_matching_tag->unpacks, p_unpack);
        }
      else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_GATE)
        {
	  nm_req_list_remove(&p_unpack->unpack.matching.p_matching_gate->unpacks, p_unpack);
        }
      else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_FULL)
	{
	  nm_req_list_remove(&p_gtag->unpacks, p_unpack);
	}
      else
        {
          NM_FATAL("unspecified matching method.\n");
        }
      if((p_pw != NULL) && (p_pw->p_trk->kind == nm_trk_large) && (p_pw->flags & NM_PW_DYNAMIC_V0))
	{
	  nm_data_copy_to(&p_unpack->data, p_pw->chunk_offset, chunk_len, p_pw->v[0].iov_base);
	}
      const struct nm_core_event_s event =
	{
	  .status = NM_STATUS_UNPACK_COMPLETED | NM_STATUS_FINALIZED,
	  .p_req = p_unpack
	};
      nm_core_status_event(p_core, &event, p_unpack);
      p_core->n_unpacks--;
      nm_trace_var(NM_TRACE_SCOPE_CORE, NM_TRACE_EVENT_VAR_N_UNPACKS, p_core->n_unpacks, NULL);
      nm_core_polling_level(p_core);
      *pp_unpack = NULL;
      /* check for out-of-order packets already received on this tag */
      while(!nm_unexpected_gtag_list_empty(&p_gtag->unexpected))
	{
	  struct nm_unexpected_s*p_unexpected = nm_unexpected_gtag_list_begin(&p_gtag->unexpected);
	  assert(p_unexpected != NULL);
	  struct nm_req_s*p_unpack = nm_unpack_find_matching(p_core, p_unexpected->p_gate, p_unexpected->seq, p_unexpected->tag);
	  if(p_unpack)
	    {
	      nm_core_unpack_unexpected(p_core, &p_unpack, p_unexpected);
	    }
	  else
	    break;
	}
    }
  else if(p_unpack->unpack.cumulated_len > p_unpack->unpack.expected_len)
    {
      NM_FATAL("copied more data than expected in unpack (unpacked = %lu; expected = %lu; chunk = %lu).\n",
	       p_unpack->unpack.cumulated_len, p_unpack->unpack.expected_len, chunk_len);
      abort();
    }
}

/** decode and manage data flags found in a matched data/short_data/rdv chunk */
static inline void nm_core_unpack_flags_decode(struct nm_req_s*p_unpack, nm_proto_t flags, nm_len_t chunk_offset, nm_len_t chunk_len)
{
  const nm_len_t chunk_end = chunk_offset + chunk_len;
  assert(p_unpack->unpack.expected_len != NM_LEN_UNDEFINED);
  if(chunk_end > p_unpack->unpack.expected_len)
    {
      NM_FATAL("received more data than expected-"
	       "   received: chunk offset = %lu; chunk len = %lu; chunk end = %lu\n"
	       "   matched request = %p; expected len = %lu; already received = %lu\n",
	       chunk_offset, chunk_len, chunk_end,
	       p_unpack, p_unpack->unpack.expected_len, p_unpack->unpack.cumulated_len);
      abort();
    }
  assert(chunk_end <= p_unpack->unpack.expected_len);
  assert(p_unpack->unpack.cumulated_len + chunk_len <= p_unpack->unpack.expected_len);
  if(flags & NM_PROTO_FLAG_LASTCHUNK)
    {
      /* Update the real size to receive */
      p_unpack->unpack.expected_len = chunk_end;
    }
  if((p_unpack->unpack.cumulated_len == 0) && (flags & NM_PROTO_FLAG_ACKREQ))
    {
      nm_core_post_ack(p_unpack->p_gate, p_unpack->tag, p_unpack->seq);
    }
}

/** insert a unexpected chunk in a sorted list */
#define nm_unexpected_insert_chunk(TYPE, P_CHUNK, P_LIST)               \
  {                                                                     \
    nm_unexpected_gtag_itor_t i = nm_unexpected_##TYPE##_list_rbegin(P_LIST); \
    while((i != NULL) && ((P_CHUNK)->seq < i->seq))                     \
      {                                                                 \
        i = nm_unexpected_##TYPE##_list_rnext(i);                       \
      }                                                                 \
    if(i)                                                               \
      {                                                                 \
        nm_unexpected_##TYPE##_list_insert_after(P_LIST, i, P_CHUNK);   \
      }                                                                 \
    else                                                                \
      {                                                                 \
        nm_unexpected_##TYPE##_list_push_front(P_LIST, P_CHUNK);        \
      }                                                                 \
  }

/** store an unexpected chunk of data (data/short_data/rdv) */
static inline void nm_unexpected_store(struct nm_core*p_core, nm_gate_t p_gate, const struct nm_header_generic_s*p_header,
				       nm_len_t chunk_offset, nm_len_t chunk_len, nm_core_tag_t tag, nm_seq_t seq,
				       struct nm_pkt_wrap_s*p_pw)
{
  if(p_pw->flags & NM_PW_BUF_RECV)
    {
      if(!(p_pw->flags & NM_PW_BUF_MIRROR))
	{
	  memcpy(p_pw->buf, p_pw->v[0].iov_base, p_pw->v[0].iov_len);
	  p_pw->flags |= NM_PW_BUF_MIRROR;
	}
	const nm_len_t header_offset = ((void*)p_header) - p_pw->v[0].iov_base;
	p_header = ((void*)p_pw->buf) + header_offset;
    }
  struct nm_unexpected_s*p_chunk = nm_unexpected_alloc();
  p_chunk->p_header = p_header;
  p_chunk->p_pw   = p_pw;
  p_chunk->p_gate = p_gate;
  p_chunk->seq    = seq;
  p_chunk->tag    = tag;
  nm_pw_ref_inc(p_pw);
  nm_profile_inc(p_core->profiling.n_unexpected);
  /* insert unexpected chunk in matching lists and cache pointers to hashtable entries */
  nm_core_lock_assert(p_core);
  struct nm_matching_wildcard_s*p_wildcard = nm_matching_wildcard_bytag(p_core, p_chunk->tag);
  p_chunk->p_wildcard = p_wildcard;
  nm_unexpected_insert_chunk(wildcard, p_chunk, &p_wildcard->unexpected);
  struct nm_matching_tag_s*p_matching_tag = nm_matching_tag_get(&p_core->tag_table, p_chunk->tag);
  p_chunk->p_matching_tag = p_matching_tag;
  nm_unexpected_insert_chunk(tag, p_chunk, &p_matching_tag->unexpected);
  struct nm_matching_gsession_s*p_gsession = nm_matching_gsession_bytag(p_gate, p_chunk->tag);
  p_chunk->p_gsession = p_gsession;
  nm_unexpected_insert_chunk(gate, p_chunk, &p_gsession->unexpected);
  struct nm_gtag_s*p_gtag = nm_gtag_get(&p_gate->tags, tag);
  p_chunk->p_gtag = p_gtag;
  nm_unexpected_insert_chunk(gtag, p_chunk, &p_gtag->unexpected);
  /* trigger event *unexpected* */
  const nm_proto_t proto_id = p_header->proto_id;
  if(proto_id & NM_PROTO_FLAG_LASTCHUNK)
    {
      p_chunk->msg_len = chunk_offset + chunk_len;
      const struct nm_core_event_s event =
	{
	  .status = NM_STATUS_UNEXPECTED,
	  .p_gate = p_gate,
	  .tag    = tag,
	  .seq    = seq,
	  .len    = chunk_offset + chunk_len
	};
      nm_core_status_event(p_core, &event, NULL);
    }
  else
    {
      p_chunk->msg_len = NM_LEN_UNDEFINED;
    }
}

void nm_core_unpack_init(struct nm_core*p_core, struct nm_req_s*p_unpack)
{
  nm_status_init(p_unpack, NM_STATUS_UNPACK_INIT);
  nm_req_list_cell_init(p_unpack);
#ifdef DEBUG
  nm_data_null_build(&p_unpack->data);
#endif
  p_unpack->flags   = NM_REQ_FLAG_UNPACK;
  p_unpack->monitor = NM_MONITOR_NULL;
  p_unpack->unpack.cumulated_len = 0;
  p_unpack->unpack.expected_len  = NM_LEN_UNDEFINED;
}

void nm_core_unpack_data(struct nm_core*p_core, struct nm_req_s*p_unpack, const struct nm_data_s*p_data)
{
  assert(p_unpack->unpack.expected_len == NM_LEN_UNDEFINED);
  p_unpack->data = *p_data;
  p_unpack->unpack.expected_len = nm_data_size(p_data);
  if(nm_status_test(p_unpack, NM_STATUS_UNPACK_DATA))
    {
      nm_core_lock(p_core);
      struct nm_unexpected_s*p_unexpected = nm_unexpected_find_matching(p_core, p_unpack);
      while(p_unexpected)
	{
	  nm_core_unpack_unexpected(p_core, &p_unpack, p_unexpected);
	  p_unexpected = p_unpack ? nm_unexpected_find_matching(p_core, p_unpack) : NULL;
	}
      nm_core_unlock(p_core);
    }
}

static inline void nm_core_unpack_match(struct nm_core*p_core, struct nm_req_s*p_unpack,
					nm_gate_t p_gate, nm_core_tag_t tag, nm_core_tag_t tag_mask, nm_seq_t seq)
{
  /* fill-in the unpack request with matching information */
  nm_status_assert(p_unpack, NM_STATUS_UNPACK_INIT);
  p_unpack->p_gate = p_gate;
  p_unpack->p_gtag = NULL;
  p_unpack->seq    = seq;
  p_unpack->tag    = tag;
  p_unpack->unpack.tag_mask = tag_mask;
  if((p_unpack->p_gate != NM_GATE_NONE) && (p_unpack->unpack.tag_mask.tag == NM_TAG_MASK_FULL))
    {
      p_unpack->flags |= NM_REQ_FLAG_MATCHING_FULL;
    }
  else if((p_unpack->p_gate == NM_GATE_NONE) && (p_unpack->unpack.tag_mask.tag == NM_TAG_MASK_FULL))
    {
      p_unpack->flags |= NM_REQ_FLAG_MATCHING_TAG;
    }
  else if((p_unpack->p_gate != NM_GATE_NONE) && (p_unpack->unpack.tag_mask.tag == NM_TAG_MASK_FULL))
    {
      p_unpack->flags |= NM_REQ_FLAG_MATCHING_GATE;
    }
  else
    {
      p_unpack->flags |= NM_REQ_FLAG_MATCHING_WILDCARD;
    }
  p_unpack->flags |= NM_REQ_FLAG_UNPACK_MATCHED;
}

/** Match an unpack request with an event.
 */
void nm_core_unpack_match_event(struct nm_core*p_core, struct nm_req_s*p_unpack, const struct nm_core_event_s*p_event)
{
  nm_core_unpack_match(p_core, p_unpack, p_event->p_gate, p_event->tag, NM_CORE_TAG_MASK_FULL, p_event->seq);
}

/** Handle an unpack request.
 */
void nm_core_unpack_match_recv(struct nm_core*p_core, struct nm_req_s*p_unpack, nm_gate_t p_gate,
			       nm_core_tag_t tag, nm_core_tag_t tag_mask)
{
  nm_core_unpack_match(p_core, p_unpack, p_gate, tag, tag_mask, NM_SEQ_NONE);
}

/** probes whether an incoming packet matched this unposted request */
int nm_core_unpack_iprobe(struct nm_core*p_core, struct nm_req_s*p_unpack)
{
  if(!(p_unpack->flags & NM_REQ_FLAG_UNPACK_MATCHED))
    {
      NM_WARN("cannot iprobe unmatched request.\n");
      return -NM_EINVAL;
    }
  if(nm_status_test(p_unpack, NM_STATUS_UNPACK_POSTED))
    {
      NM_WARN("cannot iprobe posted request.\n");
      return -NM_EINVAL;
    }
  nm_core_lock(p_core);
  if(p_unpack->flags & NM_REQ_FLAG_MATCHING_FULL)
    {
      p_unpack->p_gtag = nm_gtag_get(&p_unpack->p_gate->tags, p_unpack->tag);
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_TAG)
    {
      p_unpack->unpack.matching.p_matching_tag = nm_matching_tag_get(&p_core->tag_table, p_unpack->tag);
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_GATE)
    {
      p_unpack->unpack.matching.p_matching_gate = nm_matching_gsession_bytag(p_unpack->p_gate, p_unpack->tag);
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_WILDCARD)
    {
      p_unpack->unpack.matching.p_wildcard = nm_matching_wildcard_bytag(p_core, p_unpack->tag);
    }
  else
    {
      NM_FATAL("matching method not specified.\n");
    }
  struct nm_unexpected_s*p_unexpected = nm_unexpected_find_matching(p_core, p_unpack);
  nm_core_unlock(p_core);
  if(p_unexpected == NULL)
    {
      nm_schedule(p_core);
      return -NM_EAGAIN;
    }
  else
    return NM_ESUCCESS;
}

int nm_core_unpack_peek(struct nm_core*p_core, struct nm_req_s*p_unpack, const struct nm_data_s*p_data)
{
  if( (p_unpack->seq == NM_SEQ_NONE) ||
      (p_unpack->p_gate == NM_GATE_NONE) ||
      !(p_unpack->flags & NM_REQ_FLAG_UNPACK_MATCHED) )
    {
      NM_WARN("cannot peek unmatched request.\n");
      return -NM_EINVAL;
    }
  nm_len_t peek_len = nm_data_size(p_data);
  nm_len_t done = 0;
  nm_core_lock(p_core);
  struct nm_unexpected_s*p_chunk;
  struct nm_gtag_s*p_gtag = nm_gtag_get(&p_unpack->p_gate->tags, p_unpack->tag);
  puk_list_foreach(nm_unexpected_gtag, p_chunk, &p_gtag->unexpected)
    {
      assert(p_unpack->p_gate == p_chunk->p_gate);
      assert(nm_core_tag_match(p_chunk->tag, p_unpack->tag, p_unpack->unpack.tag_mask));
      if(p_unpack->seq == p_chunk->seq)
	{
	  const nm_header_data_generic_t*p_header = (const nm_header_data_generic_t*)p_chunk->p_header;
	  const nm_proto_t proto_id = p_chunk->p_header->proto_id & NM_PROTO_ID_MASK;
	  switch(proto_id)
	    {
	    case NM_PROTO_PKT_DATA:
	      {
		nm_len_t chunk_len    = p_header->pkt_data.data_len;
		nm_len_t chunk_offset = p_header->pkt_data.chunk_offset;
		if(chunk_offset < peek_len)
		  {
		    if(chunk_offset + chunk_len > peek_len)
		      {
			chunk_len = peek_len - chunk_offset;
		      }
		    nm_data_pkt_unpack(p_data, &p_header->pkt_data, p_chunk->p_pw, chunk_offset, chunk_len);
		    done += chunk_len;
		  }
	      }
	      break;
	    case NM_PROTO_SHORT_DATA:
	      {
		nm_len_t chunk_len    = p_header->short_data.len;
		nm_len_t chunk_offset = 0;
		const void*ptr = ((void*)p_header) + NM_HEADER_SHORT_DATA_SIZE;
		if(chunk_offset < peek_len)
		  {
		    if(chunk_offset + chunk_len > peek_len)
		      {
			chunk_len = peek_len - chunk_offset;
		      }
		    nm_data_copy_to(p_data, chunk_offset, chunk_len, ptr);
		    done += chunk_len;
		  }
	      }
	      break;
	    case NM_PROTO_DATA:
	      {
		nm_len_t chunk_len    = p_header->data.len;
		nm_len_t chunk_offset = p_header->data.chunk_offset;
		const void*ptr = (p_header->data.skip == 0xFFFF) ? (((void*)p_header) + NM_HEADER_DATA_SIZE) :
		  p_chunk->p_pw->v[0].iov_base + (p_header->data.skip + nm_header_global_v0len(p_chunk->p_pw));
		if(chunk_offset < peek_len)
		  {
		    if(chunk_offset + chunk_len > peek_len)
		      {
			chunk_len = peek_len - chunk_offset;
		      }
		    nm_data_copy_to(p_data, chunk_offset, chunk_len, ptr);
		    done += chunk_len;
		  }
	      }
	      break;
	    case NM_PROTO_RDV:
	      {
		const nm_header_ctrl_generic_t*p_ctrl_header = (const nm_header_ctrl_generic_t*)p_chunk->p_header;
		const nm_len_t chunk_offset = p_ctrl_header->rdv.chunk_offset;
		if(chunk_offset < peek_len)
		  {
		    NM_FATAL("# nmad: nm_core_unpack_peek()- not implemented for large messages. Use nm_sr_send_header() to send header eagerly.\n");
		  }
		else
		  {
		    /* rdv chunk not in peeked heder- ignore */
		  }
	      }
	      break;
	    default:
	      NM_FATAL("unknown proto_id %d", proto_id);
	      break;
	    }
	}
      if(done == peek_len)
	break;
    }
  nm_core_unlock(p_core);
  if(done == peek_len)
    return NM_ESUCCESS;
  else
    return -NM_EAGAIN;
}

/** unpack an unexpected chunk to a matched req */
static void nm_core_unpack_unexpected(struct nm_core*p_core, struct nm_req_s**pp_unpack, struct nm_unexpected_s*p_unexpected)
{
  struct nm_req_s*p_unpack = *pp_unpack;
  const void*p_header = p_unexpected->p_header;
  const nm_proto_t proto_id = p_unexpected->p_header->proto_id & NM_PROTO_ID_MASK;
  switch(proto_id)
    {
    case NM_PROTO_PKT_DATA:
      {
	const struct nm_header_pkt_data_s*h = p_header;
	nm_pkt_data_handler(p_core, p_unpack->p_gate, pp_unpack, h, p_unexpected->p_pw);
      }
      break;
    case NM_PROTO_SHORT_DATA:
      {
	const struct nm_header_short_data_s*h = p_header;
	nm_short_data_handler(p_core, p_unpack->p_gate, pp_unpack, h, p_unexpected->p_pw);
      }
      break;
    case NM_PROTO_DATA:
      {
	const struct nm_header_data_s*h = p_header;
	nm_small_data_handler(p_core, p_unpack->p_gate, pp_unpack, h, p_unexpected->p_pw);
      }
      break;
    case NM_PROTO_RDV:
      {
	const struct nm_header_ctrl_rdv_s*h = p_header;
	nm_rdv_handler(p_core, p_unpack->p_gate, p_unpack, h, p_unexpected->p_pw);
      }
      break;
    }
  /* Decrement the packet wrapper reference counter. If no other
     chunks are still in use, the pw will be destroyed. */
  nm_pw_ref_dec(p_unexpected->p_pw);
  nm_unexpected_wildcard_list_remove(&p_unexpected->p_wildcard->unexpected, p_unexpected);
  nm_unexpected_tag_list_remove(&p_unexpected->p_matching_tag->unexpected, p_unexpected);
  nm_unexpected_gate_list_remove(&p_unexpected->p_gsession->unexpected, p_unexpected);
  nm_unexpected_gtag_list_remove(&p_unexpected->p_gtag->unexpected, p_unexpected);
  nm_unexpected_free(nm_unexpected_allocator, p_unexpected);
}

/** Store the unpack request. */
void nm_core_unpack_submit(struct nm_core*p_core, struct nm_req_s*p_unpack, nm_req_flag_t flags)
{
  nm_status_assert(p_unpack, NM_STATUS_UNPACK_INIT);
  assert(p_unpack->flags & NM_REQ_FLAG_UNPACK_MATCHED);

  nm_status_add(p_unpack, NM_STATUS_UNPACK_POSTED);
  nm_core_lock(p_core);
  p_unpack->unpack.req_seq = p_core->unpack_seq;
  p_core->unpack_seq++;
  assert(p_unpack->unpack.tag_mask.hashcode == NM_CORE_TAG_HASH_FULL);
  if(p_unpack->flags & NM_REQ_FLAG_MATCHING_FULL)
    {
      p_unpack->p_gtag = nm_gtag_get(&p_unpack->p_gate->tags, p_unpack->tag);
      nm_req_list_push_back(&p_unpack->p_gtag->unpacks, p_unpack);
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_TAG)
    {
      p_unpack->unpack.matching.p_matching_tag = nm_matching_tag_get(&p_core->tag_table, p_unpack->tag);
      nm_req_list_push_back(&p_unpack->unpack.matching.p_matching_tag->unpacks, p_unpack);
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_GATE)
    {
      p_unpack->unpack.matching.p_matching_gate = nm_matching_gsession_bytag(p_unpack->p_gate, p_unpack->tag);
      nm_req_list_push_back(&p_unpack->unpack.matching.p_matching_gate->unpacks, p_unpack);
    }
  else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_WILDCARD)
    {
      p_unpack->unpack.matching.p_wildcard = nm_matching_wildcard_bytag(p_core, p_unpack->tag);
      nm_req_list_push_back(&p_unpack->unpack.matching.p_wildcard->unpacks, p_unpack);
    }
  else
    {
      NM_FATAL("matching method not specified.\n");
    }
  nm_profile_inc(p_core->profiling.n_unpacks);
  p_core->n_unpacks++;
  nm_trace_var(NM_TRACE_SCOPE_CORE, NM_TRACE_EVENT_VAR_N_UNPACKS, p_core->n_unpacks, NULL);
#ifdef NMAD_PROFILE
  if(p_core->n_unpacks > p_core->profiling.max_unpacks)
    p_core->profiling.max_unpacks = p_core->n_unpacks;
#endif /* NMAD_PROFILE */
  nm_core_polling_level(p_core);
  struct nm_unexpected_s*p_unexpected = nm_unexpected_find_matching(p_core, p_unpack);
  if(p_unexpected && (p_unpack->unpack.expected_len == NM_LEN_UNDEFINED))
    {
      /* data spec still undefined- fire data event */
      const struct nm_core_event_s event =
	{
	  .status = NM_STATUS_UNPACK_DATA,
	  .p_req = p_unpack
	};
      nm_core_status_event(p_core, &event, p_unpack);
    }
  else
    {
      while(p_unexpected)
	{
	  /* data is already here (at least one chunk)- process all matching chunks */
	  nm_core_unpack_unexpected(p_core, &p_unpack, p_unexpected);
	  p_unexpected = p_unpack ? nm_unexpected_find_matching(p_core, p_unpack) : NULL;
	}
    }
  nm_core_unlock(p_core);
}

int nm_core_iprobe(struct nm_core*p_core,
		   nm_gate_t p_gate, nm_core_tag_t tag, nm_core_tag_t tag_mask,
		   nm_gate_t *pp_out_gate, nm_core_tag_t*p_out_tag, nm_len_t*p_out_size)
{
  int rc = -NM_EAGAIN;
  struct nm_unexpected_s*p_chunk;
  nm_core_lock(p_core);
  struct nm_matching_wildcard_s*p_wildcard = nm_matching_wildcard_bytag(p_core, tag);
  puk_list_foreach(nm_unexpected_wildcard, p_chunk, &p_wildcard->unexpected)
    {
      struct nm_gtag_s*p_gtag = p_chunk->p_gtag;
      const nm_seq_t next_seq = nm_seq_next(p_gtag->recv_seq_number);
      if(((p_gate == p_chunk->p_gate) || (p_gate == NM_ANY_GATE)) && /* gate matches */
	 nm_core_tag_match(p_chunk->tag, tag, tag_mask) && /* tag matches */
	 (p_chunk->seq == next_seq) /* seq number matches */ )
	{
	  const void*header = p_chunk->p_header;
	  const nm_proto_t proto_id = (*(nm_proto_t*)header) & NM_PROTO_ID_MASK;
	  const nm_proto_t proto_flags = (*(nm_proto_t*)header) & NM_PROTO_FLAG_MASK;
	  int matches = 0;
	  nm_len_t len = -1;
	  switch(proto_id)
	    {
	    case NM_PROTO_PKT_DATA:
	      {
		const struct nm_header_pkt_data_s*h = header;
		if(proto_flags & NM_PROTO_FLAG_LASTCHUNK)
		  {
		    len = h->data_len + h->chunk_offset;
		    matches = 1;
		  }
	      }
	      break;
	    case NM_PROTO_SHORT_DATA:
	      {
		const struct nm_header_short_data_s *h = header;
		len = h->len;
		matches = 1;
	      }
	      break;
	    case NM_PROTO_DATA:
	      {
		const struct nm_header_data_s *h = header;
		if(proto_flags & NM_PROTO_FLAG_LASTCHUNK)
		  {
		    len = h->len + h->chunk_offset;
		    matches = 1;
		  }
	      }
	      break;
	    case NM_PROTO_RDV:
	      {
		const struct nm_header_ctrl_rdv_s *rdv = header;
		if(proto_flags & NM_PROTO_FLAG_LASTCHUNK)
		  {
		    len = rdv->len + rdv->chunk_offset;
		    matches = 1;
		  }
	      }
	      break;
	    }
	  if(matches)
	    {
	      if(pp_out_gate)
		*pp_out_gate = p_chunk->p_gate;
	      if(p_out_tag)
		*p_out_tag = p_chunk->tag;
	      if(p_out_size)
		*p_out_size = len;
	      rc = NM_ESUCCESS;
	      goto out;
	    }
	}
    }
  *pp_out_gate = NM_ANY_GATE;
 out:
  nm_core_unlock(p_core);
  return rc;
}

int nm_core_unpack_cancel(struct nm_core*p_core, struct nm_req_s*p_unpack)
{
  int rc = NM_ESUCCESS;
  nm_core_lock(p_core);
  if(nm_status_test(p_unpack, NM_STATUS_UNPACK_CANCELLED))
    {
      /* received already canacelled */
      rc = -NM_ECANCELED;
    }
  else if(nm_status_test(p_unpack, NM_STATUS_UNPACK_COMPLETED))
    {
      /* receive already completed */
      rc = -NM_EALREADY;
    }
  else if(p_unpack->seq == NM_SEQ_NONE)
    {
      if(p_unpack->flags & NM_REQ_FLAG_MATCHING_WILDCARD)
	{
	  nm_req_list_remove(&p_unpack->unpack.matching.p_wildcard->unpacks, p_unpack);
	}
      else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_TAG)
        {
	  nm_req_list_remove(&p_unpack->unpack.matching.p_matching_tag->unpacks, p_unpack);
        }
      else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_GATE)
        {
	  nm_req_list_remove(&p_unpack->unpack.matching.p_matching_gate->unpacks, p_unpack);
        }
      else if(p_unpack->flags & NM_REQ_FLAG_MATCHING_FULL)
	{
	  nm_req_list_remove(&p_unpack->p_gtag->unpacks, p_unpack);
	}
      else
        {
          NM_FATAL("unspecified matching method.\n");
        }
      nm_core_polling_level(p_core);
      const struct nm_core_event_s event =
	{
	  .status = NM_STATUS_UNPACK_CANCELLED | NM_STATUS_FINALIZED,
	  .p_req = p_unpack
	};
      nm_core_status_event(p_core, &event, p_unpack);
      rc = NM_ESUCCESS;
    }
  else
    {
      /* receive is already in progress- too late to cancel */
      rc = -NM_EINPROGRESS;
    }
  nm_core_unlock(p_core);
  return rc;
}

/** checks whether an unpack request is ready to receive data, store chunk as unexpected otherwise */
static inline int nm_core_unpack_ready_to_receive(struct nm_core*p_core, struct nm_req_s*p_unpack,
						  nm_gate_t p_gate, const struct nm_header_generic_s*p_header,
						  nm_len_t chunk_offset, nm_len_t chunk_len, nm_core_tag_t tag, nm_seq_t seq,
						  struct nm_pkt_wrap_s*p_pw)
{
  if((p_unpack != NULL) &&
     (!(nm_status_test(p_unpack, NM_STATUS_UNPACK_CANCELLED))) &&
     (p_unpack->unpack.expected_len != NM_LEN_UNDEFINED))
    {
      return 1;
    }
  nm_unexpected_store(p_core, p_gate, p_header, chunk_offset, chunk_len, tag, seq, p_pw);
  if((p_unpack != NULL) && (p_unpack->unpack.expected_len == NM_LEN_UNDEFINED))
    {
      /* data spec still undefined- fire data event */
      const struct nm_core_event_s event =
	{
	  .status = NM_STATUS_UNPACK_DATA,
	  .p_req = p_unpack
	};
      nm_core_status_event(p_core, &event, p_unpack);
    }
  return 0;
}

/** Process a packed data request (NM_PROTO_PKT_DATA)- p_unpack may be NULL (unexpected)
 */
static void nm_pkt_data_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s**pp_unpack,
				const struct nm_header_pkt_data_s*h, struct nm_pkt_wrap_s *p_pw)
{
  const nm_len_t chunk_len = h->data_len;
  const nm_len_t chunk_offset = h->chunk_offset;
  if(nm_core_unpack_ready_to_receive(p_core, *pp_unpack,
				     p_gate, (void*)h, chunk_offset, chunk_len, h->tag_id, h->seq, p_pw))
    {
      assert(nm_status_test(*pp_unpack, NM_STATUS_UNPACK_POSTED));
      nm_core_unpack_flags_decode(*pp_unpack, h->proto_id & NM_PROTO_FLAG_MASK, chunk_offset, chunk_len);
      nm_data_pkt_unpack(&(*pp_unpack)->data, h, p_pw, chunk_offset, chunk_len);
      nm_core_unpack_check_completion(p_core, p_pw, pp_unpack, chunk_len);
    }
}

/** Process a short data request (NM_PROTO_SHORT_DATA)- p_unpack may be NULL (unexpected)
 */
static void nm_short_data_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s**pp_unpack,
				  const struct nm_header_short_data_s*h, struct nm_pkt_wrap_s *p_pw)
{
  const nm_len_t chunk_len = h->len;
  const nm_len_t chunk_offset = 0;
  if(nm_core_unpack_ready_to_receive(p_core, *pp_unpack,
				     p_gate, (void*)h, chunk_offset, chunk_len, h->tag_id, h->seq, p_pw))
    {
      const nm_proto_t flags = NM_PROTO_FLAG_LASTCHUNK;
      const void*ptr = ((void*)h) + NM_HEADER_SHORT_DATA_SIZE;
      nm_core_unpack_flags_decode(*pp_unpack, flags, chunk_offset, chunk_len);
      nm_data_copy_to(&(*pp_unpack)->data, chunk_offset, chunk_len, ptr);
      nm_core_unpack_check_completion(p_core, p_pw, pp_unpack, chunk_len);
    }
}

/** Process a small data request (NM_PROTO_DATA)- p_unpack may be NULL (unexpected)
 */
static void nm_small_data_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s**pp_unpack,
				  const struct nm_header_data_s*h, struct nm_pkt_wrap_s*p_pw)
{
  const void*ptr = (h->skip == 0xFFFF) ? (((void*)h) + NM_HEADER_DATA_SIZE) :
    p_pw->v[0].iov_base + (h->skip + nm_header_global_v0len(p_pw));
  assert(p_pw->v_nb == 1);
  const nm_len_t chunk_len = h->len;
  const nm_len_t chunk_offset = h->chunk_offset;
  if(nm_core_unpack_ready_to_receive(p_core, *pp_unpack,
				     p_gate, (void*)h, chunk_offset, chunk_len, h->tag_id, h->seq, p_pw))
    {
      nm_core_unpack_flags_decode(*pp_unpack, h->proto_id & NM_PROTO_FLAG_MASK, chunk_offset, chunk_len);
      nm_data_copy_to(&(*pp_unpack)->data, chunk_offset, chunk_len, ptr);
      nm_core_unpack_check_completion(p_core, p_pw, pp_unpack, chunk_len);
    }
}

/** Process a received rendez-vous request (NM_PROTO_RDV)-
 * either p_unpack may be NULL (storing unexpected) or p_pw (unpacking unexpected)
 */
static void nm_rdv_handler(struct nm_core*p_core, nm_gate_t p_gate, struct nm_req_s*p_unpack,
			   const struct nm_header_ctrl_rdv_s*h, struct nm_pkt_wrap_s *p_pw)
{
  const nm_len_t chunk_len = h->len;
  const nm_len_t chunk_offset = h->chunk_offset;
  if(nm_core_unpack_ready_to_receive(p_core, p_unpack,
				     p_gate, (void*)h, chunk_offset, chunk_len, h->tag_id, h->seq, p_pw))
    {
      nm_profile_inc(p_core->profiling.n_rdvs);
      assert(p_unpack->p_gate != NULL);
      assert(!nm_data_isnull(&p_unpack->data));
      nm_core_unpack_flags_decode(p_unpack, h->proto_id & NM_PROTO_FLAG_MASK, chunk_offset, chunk_len);
      struct nm_pkt_wrap_s*p_large_pw = nm_pw_alloc_noheader(p_core);
      p_large_pw->p_unpack     = p_unpack;
      p_large_pw->length       = chunk_len;
      p_large_pw->chunk_offset = chunk_offset;
      p_large_pw->p_data       = &p_unpack->data;
      p_large_pw->p_gate       = p_gate;
      nm_pkt_wrap_list_push_back(&p_gate->pending_large_recv, p_large_pw);
      nm_gate_set_active(p_gate);
    }
}

/** Process a complete rendez-vous ready-to-receive request.
 */
static void nm_rtr_handler(struct nm_pkt_wrap_s*p_rtr_pw, const struct nm_header_ctrl_rtr_s*header)
{
  const nm_core_tag_t tag     = header->tag_id;
  const nm_seq_t seq          = header->seq;
  const nm_len_t chunk_offset = header->chunk_offset;
  const nm_len_t chunk_len    = header->chunk_len;
  nm_gate_t p_gate      = p_rtr_pw->p_gate;
  struct nm_core*p_core = p_gate->p_core;
  struct nm_pkt_wrap_s*p_large_pw = NULL, *p_pw_save;
  nm_core_lock_assert(p_core);
  puk_list_foreach_safe(nm_pkt_wrap, p_large_pw, p_pw_save, &p_gate->pending_large_send)
    {
      /* this is a large pw with rdv- it must contain a single req chunk */
      assert(nm_req_chunk_list_size(&p_large_pw->req_chunks) == 1);
      struct nm_req_chunk_s*p_req_chunk = nm_req_chunk_list_begin(&p_large_pw->req_chunks);
      struct nm_req_s*p_pack = p_req_chunk->p_req;
      NM_TRACEF("Searching the pw corresponding to the ack - cur_seq = %d - cur_offset = %d\n",
		p_pack->seq, p_large_pw->chunk_offset);
      if((p_pack->seq == seq) && nm_core_tag_eq(p_pack->tag, tag) && (p_large_pw->chunk_offset == chunk_offset))
	{
	  nm_pkt_wrap_list_remove(&p_gate->pending_large_send, p_large_pw);
	  if(chunk_len < p_large_pw->length)
	    {
	      /* ** partial RTR- split the packet  */
	      /* assert ack is partial */
	      assert(chunk_len > 0 && chunk_len < p_large_pw->length);
	      /* create a new pw with the remaining data */
	      struct nm_pkt_wrap_s*p_pw2 = nm_pw_alloc_noheader(p_core);
	      p_pw2->p_drv    = p_large_pw->p_drv;
	      p_pw2->trk_id   = p_large_pw->trk_id;
	      p_pw2->p_gate   = p_gate;
	      p_pw2->p_trk    = p_large_pw->p_trk;
	      p_pw2->p_unpack = NULL;
	      struct nm_req_chunk_s*p_req_chunk2 = nm_req_chunk_alloc(p_core);
	      nm_req_chunk_init(p_req_chunk2, p_req_chunk->p_req, NM_LEN_UNDEFINED, p_req_chunk->chunk_len - chunk_len);
	      nm_req_chunk_list_push_back(&p_pw2->req_chunks, p_req_chunk2);
	      p_req_chunk->chunk_len = chunk_len; /* truncate the contrib */
	      /* populate p_pw2 iovec */
	      nm_pw_split_data(p_large_pw, p_pw2, chunk_len);
	      nm_pkt_wrap_list_push_front(&p_gate->pending_large_send, p_pw2);
	    }
	  /* send the data */
	  struct nm_trk_s*p_trk = nm_trk_get_by_index(p_gate, header->trk_id);
	  if(p_trk->kind == nm_trk_large)
	    {
	      nm_core_post_send(p_large_pw, p_gate, header->trk_id);
	    }
	  else
	    {
	      /* rdv eventually accepted on trk#0- rollback and repack */
	      assert(p_large_pw->p_data != NULL);
	      nm_pw_free(p_core, p_large_pw);
	      struct nm_req_chunk_s*p_req_chunk0 = nm_req_chunk_malloc(&p_core->req_chunk_allocator);
	      nm_req_chunk_init(p_req_chunk0, p_pack, chunk_offset, chunk_len);
	      nm_req_chunk_list_push_back(&p_gate->req_chunk_list, p_req_chunk0);
              nm_gate_set_active(p_gate);
	    }
	  return;
	}
    }
  NM_FATAL("FATAL- cannot find matching packet for received RTR: seq = %d; offset = %lu\n", seq, chunk_offset);
  abort();
}
/** Process an acknowledgement.
 */
static void nm_ack_handler(struct nm_pkt_wrap_s *p_ack_pw, const struct nm_header_ctrl_ack_s*header)
{
  struct nm_core*p_core = p_ack_pw->p_gate->p_core;
  const nm_core_tag_t tag = header->tag_id;
  const nm_seq_t seq = header->seq;
  struct nm_req_s*p_pack = NULL;
  struct nm_gtag_s*p_gtag = nm_gtag_get(&p_ack_pw->p_gate->tags, tag);

  puk_list_foreach(nm_req, p_pack, &p_gtag->pending_packs)
    {
      assert(nm_core_tag_eq(p_pack->tag, tag));
      if(p_pack->seq == seq)
	{
	  const int finalized = nm_status_test(p_pack, NM_STATUS_PACK_COMPLETED);
	  nm_core_lock_assert(p_core);
	  const struct nm_core_event_s event =
	    {
	      .status = NM_STATUS_ACK_RECEIVED | ( finalized ? NM_STATUS_FINALIZED : 0),
	      .p_req = p_pack
	    };
	  if(finalized)
	    {
	      nm_req_list_remove(&p_gtag->pending_packs, p_pack);
	      p_core->n_packs--;
              nm_trace_var(NM_TRACE_SCOPE_CORE, NM_TRACE_EVENT_VAR_N_PACKS, p_core->n_packs, NULL);
	    }
	  nm_core_status_event(p_core, &event, p_pack);
	  nm_core_polling_level(p_core);
	  return;
	}
    }
}

/** decode one chunk of headers.
 * @returns the number of processed bytes in global header,
 */
int nm_decode_header_chunk(struct nm_core*p_core, const void*ptr, struct nm_pkt_wrap_s*p_pw, nm_gate_t p_gate)
{
  int rc = 0;
  const nm_proto_t proto_id   = (*(const nm_proto_t*)ptr) & NM_PROTO_ID_MASK;
  const nm_proto_t proto_flag = (*(const nm_proto_t*)ptr) & NM_PROTO_FLAG_MASK;

  assert(proto_id != 0);
  switch(proto_id)
    {
    case NM_PROTO_PKT_DATA:
      {
	const struct nm_header_pkt_data_s*h = ptr;
	struct nm_req_s*p_unpack = nm_unpack_find_matching(p_core, p_gate, h->seq, h->tag_id);
	nm_pkt_data_handler(p_core, p_gate, &p_unpack, h, p_pw);
	rc = h->hlen;
      }
      break;

    case NM_PROTO_SHORT_DATA:
      {
	const struct nm_header_short_data_s*h = ptr;
	struct nm_req_s*p_unpack = nm_unpack_find_matching(p_core, p_gate, h->seq, h->tag_id);
	nm_short_data_handler(p_core, p_gate, &p_unpack, h, p_pw);
	rc = NM_HEADER_SHORT_DATA_SIZE + h->len;
      }
      break;

    case NM_PROTO_DATA:
      {
	const struct nm_header_data_s*h = ptr;
	rc = NM_HEADER_DATA_SIZE;
	if(h->skip == 0xFFFF)
	  {
	    const nm_len_t size = (proto_flag & NM_PROTO_FLAG_ALIGNED) ? nm_aligned(h->len) : h->len;
	    rc += size;
	  }
	struct nm_req_s*p_unpack = nm_unpack_find_matching(p_core, p_gate, h->seq, h->tag_id);
	nm_small_data_handler(p_core, p_gate, &p_unpack, h, p_pw);
      }
      break;

    case NM_PROTO_RDV:
      {
	const struct nm_header_ctrl_rdv_s*h = ptr;
	rc = NM_HEADER_CTRL_SIZE;
	struct nm_req_s*p_unpack = nm_unpack_find_matching(p_core, p_gate, h->seq, h->tag_id);
	nm_rdv_handler(p_core, p_gate, p_unpack, h, p_pw);
      }
      break;

    case NM_PROTO_RTR:
      {
	const struct nm_header_ctrl_rtr_s*h = ptr;
	rc = NM_HEADER_CTRL_SIZE;
	nm_rtr_handler(p_pw, h);
      }
      break;

    case NM_PROTO_ACK:
      {
	const struct nm_header_ctrl_ack_s*h = ptr;
	rc = NM_HEADER_CTRL_SIZE;
	nm_ack_handler(p_pw, h);
      }
      break;

    case NM_PROTO_STRAT:
      {
	const struct puk_receptacle_NewMad_Strategy_s*strategy = &p_gate->strategy_receptacle;
	const struct nm_header_strat_s*h = ptr;
	if(strategy->driver->proto)
	  {
	    (*strategy->driver->proto)(strategy->_status, p_gate, p_pw, ptr, h->size);
	  }
	else
	  {
	    NM_FATAL("strategy cannot process NM_PROTO_STRAT.\n");
	  }
	rc = h->size;
      }
      break;

    default:
      NM_FATAL("received header with invalid proto_id 0x%02X\n", proto_id);
      break;
    }
  return rc;
}


/** Process a complete incoming request.
 */
void nm_pw_process_complete_recv(struct nm_core*p_core, struct nm_pkt_wrap_s*p_pw)
{
  nm_drv_t const p_drv = p_pw->p_drv;
  nm_core_lock_assert(p_core);
  nm_profile_inc(p_core->profiling.n_pw_in);
  /* check error status */
  if(p_pw->flags & NM_PW_CANCELED)
    {
      nm_pw_ref_dec(p_pw);
      return;
    }
  /* clear the input request field */
  assert(p_pw->p_trk);
  assert(p_pw->p_trk->p_pw_recv == p_pw);
  p_pw->p_trk->p_pw_recv = NULL;
  assert((p_drv->p_pw_recv_any == p_pw) || (p_drv->p_pw_recv_any == NULL));
  p_drv->p_pw_recv_any = NULL;
  nm_gate_t const p_gate = p_pw->p_gate;
  assert(p_gate != NULL);
  if(p_pw->flags & NM_PW_CLOSED)
    {
      nm_pw_ref_dec(p_pw);
      nm_gate_disconnected(p_core, p_gate, p_drv);
      return;
    }

  if(p_pw->p_trk->kind == nm_trk_small)
    {
      /* ** Small packets - track #0 *********************** */
      struct iovec*const v0 = p_pw->v;
      const nm_len_t v0len = nm_header_global_v0len(p_pw);
      const void*ptr = v0->iov_base + sizeof(struct nm_header_global_s);
      do
	{
	  /* Iterate over header chunks */
	  assert(ptr < v0->iov_base + v0->iov_len);
	  nm_len_t done = nm_decode_header_chunk(p_core, ptr, p_pw, p_gate);
	  ptr += done;
	}
      while(ptr < v0->iov_base + v0len);
      if(p_pw->flags & NM_PW_BUF_RECV)
	{
	  const struct puk_receptacle_NewMad_minidriver_s*r = &p_pw->p_trk->receptacle;
	  (*r->driver->recv_buf_release)(r->_status);
	  if(p_pw->flags & NM_PW_BUF_MIRROR)
	    {
	      p_pw->v[0].iov_base = p_pw->buf;
	    }
	  else
	    {
#ifdef DEBUG
	      p_pw->v[0].iov_base = NULL;
	      p_pw->v[0].iov_len = 0;
#endif /* DEBUG*/
	    }
	}
      /* refill recv on trk #0 */
      nm_drv_refill_recv(p_drv, p_gate);
    }
  else if(p_pw->p_trk->kind == nm_trk_large)
    {
      /* ** Large packet - track #1 ************************ */
      struct nm_req_s*p_unpack = p_pw->p_unpack;
      const nm_len_t len = p_pw->length;
      /* ** Large packet, data received directly in its final destination */
      nm_core_unpack_check_completion(p_core, p_pw, &p_unpack, len);
    }
  nm_pw_ref_dec(p_pw);
  nm_strat_rdv_accept(p_gate);
  /* Hum... Well... We're done guys! */
}
